     /*-                                                                            +
      * Copyright (c) 2001-2008 Scott C. Klement                                    +
      * All rights reserved.                                                        +
      *                                                                             +
      * Redistribution and use in source and binary forms, with or without          +
      * modification, are permitted provided that the following conditions          +
      * are met:                                                                    +
      * 1. Redistributions of source code must retain the above copyright           +
      *    notice, this list of conditions and the following disclaimer.            +
      * 2. Redistributions in binary form must reproduce the above copyright        +
      *    notice, this list of conditions and the following disclaimer in the      +
      *    documentation and/or other materials provided with the distribution.     +
      *                                                                             +
      * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ''AS IS'' AND      +
      * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE       +
      * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  +
      * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE     +
      * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL  +
      * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS     +
      * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       +
      * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT  +
      * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY   +
      * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF      +
      * SUCH DAMAGE.                                                                +
      *                                                                             +
      */                                                                            +

     /*
      * HTTPAPIR4 -- Hypertext transfer protocol API
      *
      *   A suite of routines for doing HTTP from ILE programs.
      *
      *   Requires:
      *       ILE RPG/400 licensed program from IBM.  V4R2 or later.
      *
      *       Optional TLS/SSL capabilities require V4R5 or later.
      *       and require the HTTP server, Client Encryption and
      *       Digital Certificate Manger Licensed programs.
      *
      *       See the README member, included in this package for details.
      *
      *   To Compile:
      *      - Edit the CONFIG_H member to set options.
      *      CRTCLPGM INSTALL SRCFILE(libhttp/QCLSRC)
      *      CALL INSTALL
      *
      *  To Update this module (HTTPAPIR4) without running
      *     the whole install procedure:
      *
      *      CRTRPGMOD HTTPAPIR4 SRCFILE(LIBHTTP/QRPGLESRC) DBGVIEW(*LIST)
      *      UPDSRVPGM SRVPGM(LIBHTTP/HTTPAPIR4) MODULE(HTTPAPIR4) +
      *                EXPORT(*CURRENT)
      */

     H COPYRIGHT('Copyright (c) 2001-2008 Scott C. Klement.  All rights +
     H reserved.  A member called "LICENSE" was included with this +
     H distribution, and contains important license information.')

      /if defined(HAVE_SRCSTMT_NODEBUGIO)
     H OPTION(*NOSHOWCPY: *SRCSTMT: *NODEBUGIO)
      /endif
     H NOMAIN

      /copy QRPGLESRC,HTTPAPI_H
      /copy QRPGLESRC,PRIVATE_H
      /copy QRPGLESRC,HEADER_H
      /copy QRPGLESRC,ERRNO_H
      /copy QRPGLESRC,IFSIO_H
      /copy QRPGLESRC,COMM_H

      /if defined(HTTP_USE_CCSID)
     D CCSID_OR_CP     S             10I 0 inz(O_CCSID)
      /else
     D CCSID_OR_CP     S             10I 0 inz(O_CODEPAGE)
      /endif

      ***  Local procedures ***

     D do_get          PR            10I 0
     D  peSaveProc                     *   value procptr
     D  peComm                         *   value
     D  peFile                       10I 0 value
     D  peTimeout                    10I 0 value
     D  peAbsPath                 32767A   varying const options(*varsize)
     D  peHost                      256A   const
     D  peModTime                      Z   options(*omit)
     D  peUserAgent                  64A   options(*omit)
     D  peContentType                64A   options(*omit)
     D  peSoapAction                 64A   options(*omit)
     D  pePort                       10I 0 value
     D  peSecure                      1N   value
     D  peServ                       32A   const

     D do_post         PR            10I 0
     D  peSaveProc                     *   value procptr
     D  peSendProc                     *   value procptr
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peComm                         *   value
     D  peFile                       10I 0 value
     D  peTimeout                    10I 0 value
     D  peAbsPath                 32767A   const varying options(*varsize)
     D  peHost                      256A   const
     D  peUserAgent                  64A   options(*omit)
     D  peContentType                64A   options(*omit)
     D  peSoapAction                 64A   options(*omit)
     D  pePostProc                     *   value procptr
     D  pePostFD                     10I 0 value
     D  pePort                       10I 0 value
     D  peSecure                      1N   value
     D  peServ                       32A   const

     D SendReq         PR            10I 0
     D   peComm                        *   value
     D   peData                        *   value
     D   peDataLen                   10I 0 value
     D   peTimeout                   10I 0 value

     D recvresp        PR            10I 0
     D   peComm                        *   value
     D   peRespChain              32767A   varying options(*varsize)
     D   peRespLen                   10I 0 value
     D   peTimeOut                   10P 3 value
     D   peUse100                     1N   const

     D recvdoc         PR            10I 0
     D   peComm                        *   value
     D   peProcedure                   *   value procptr
     D   peFD                        10I 0 value
     D   peTimeout                   10I 0 value
     D   peCLen                      10U 0 value
     D   peUseCL                      1N   const

     D recvchunk       PR            10I 0
     D   peComm                        *   value
     D   peProcedure                   *   value procptr
     D   peFD                        10I 0 value
     D   peTimeout                   10I 0 value

     D SendDoc         PR            10I 0
     D  peComm                         *   value
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peTimeout                    10I 0 value
     D  peUnused1                      *   value procptr
     D  peUnused2                    10I 0 value

     D SendRaw         PR            10I 0
     D  peComm                         *   value
     D  peUnused1                      *   value
     D  peDataSize                   10I 0 value
     D  peTimeout                    10I 0 value
     D  pePostProc                     *   value procptr
     D  pePostFD                     10I 0 value

     d get_chunk_size  PR            10I 0
     d   peComm                        *   value
     d   peTimeout                   10I 0 value

     D interpret_auth  PR
     D   peRespChain               2048A   const
     D   peKwdPos                    10I 0 value

     D interpret_proxy_auth...
     D                 PR
     D   peRespChain               2048A   const
     D   peKwdPos                    10I 0 value

     D mkdigest        PR         32767A   varying
     D   peMethod                     4A   const
     D   peURI                    32767A   varying const options(*varsize)

     D parse_resp_chain...
     D                 PR            10I 0
     D  peRespChain               32767A   varying const
     D  peRC                         10I 0
     D  peTE                         32A
     D  peCLen                       10u 0
     D  peUseCL                       1N
     D  peAuthErr                     1N
     D  peProxyAuthErr...
     D                                1N
     D  peHost                      256A   varying const
     D  pePath                      256A   varying const

     D setUrlAuth      PR
     D   peUsername                  80A   const
     D   pePasswd                  1024A   const

     D upload_sts      PR                  ExtProc(wkUplProc)
     D   peBytesSent                 10U 0 value
     D   peBytesTot                  10U 0 value
     D   peUserData                    *   value

     D download_sts    PR                  ExtProc(wkDwnlProc)
     D   peBytesSent                 10U 0 value
     D   peBytesTot                  10U 0 value
     D   peUserData                    *   value

     D addl_headers    PR                  ExtProc(wkAddHdrProc)
     D   peAddlHdrs                1024A   varying
     D   peUserData                    *   value

     D parse_hdrs      PR                  ExtProc(wkParseHdrProc)
     D   peHdrData                 2048A   const
     D   peUserData                    *   value

     D parse_hdr_long  PR                  ExtProc(wkParseHdrLong)
     D   peHdrData                32767A   const varying
     D   peUserData                    *   value

     D proxy_tunnel    PR            10I 0
     D   peComm                        *   value
     D   peServ                      32a   const
     D   peHost                     256a   const
     D   pePort                      10i 0 value
     D   peTimeout                   10i 0 value


      ***  Global Constants  ***

     D CRLF            C                   CONST(x'0d25')
     D upper           C                   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     D lower           C                   'abcdefghijklmnopqrstuvwxyz'

      ***  Global Variables  ***

     D wkRedirLoc      s           1024A   varying inz('')

     D dsAuth          DS
     D   dsAuthType                   1A   inz(HTTP_AUTH_NONE)
     D   dsAuthBasic                  1N   inz(*OFF)
     D   dsAuthDigest                 1N   inz(*OFF)
     D   dsAuthRealm                124A
     D   dsAuthNonce                128A
     D   dsAuthOpaque               128A
     D   dsAuthQOP                   32A
     D   dsAuthCnonce                16A
     D   dsAuthNC                     7P 0 inz(0)
     D   dsAuthUser                  80A
     D   dsAuthPasswd              1024A
     D   dsAuthStr                 1476A   varying inz('')

     D dsProxyAuth     DS
     D   dsProxyAuthType...
     D                                1A   inz(HTTP_AUTH_NONE)
     D   dsProxyAuthBasic...
     D                                1N   inz(*OFF)
     D   dsProxyAuthRealm...
     D                              124A
     D   dsProxyAuthUser...
     D                               80A
     D   dsProxyAuthPasswd...
     D                             1024A
     D   dsProxyAuthStr...
     D                             1476A   varying inz('')

     D dsProxy         DS
     D   dsProxyHost               2048A   inz(*blanks)
     D   dsProxyPort                 10I 0 inz(*zeros)
     D   dsProxyTun                   1N   inz(*off)

     D wkSaveAuth      s                   like(dsAuth)

     D wkUplProc       S               *   procptr inz(*NULL)
     D wkUplUdata      S               *   inz(*NULL)
     D wkDwnlProc      S               *   procptr inz(*NULL)
     D wkDwnlUdata     S               *   inz(*NULL)
     D wkAddHdrProc    S               *   procptr inz(*NULL)
     D wkAddHdrData    S               *   inz(*NULL)
     D wkParseHdrProc  S               *   procptr inz(*NULL)
     D wkParseHdrData  S               *   inz(*NULL)
     D wkParseHdrLong  S               *   procptr inz(*NULL)
     D wkParseHdrLongData...
     D                 S               *   inz(*NULL)

     D wk100time       S             10P 3 inz(0.5)
     D use_cookies     S              1N   inz(HTTP_COOKIE_DEFAULT)

     D wkDebugNoClose  S               N   inz(*OFF)


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  Close HTTP connection
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_close      B                   export
     D http_close      PI            10I 0
     D  peSock                       10I 0 value
     D  peComm                         *   value

     c                   callp     debug_msg('http_close(): entered')

     c                   if        %parms < 2
     c                   callp     SetError(HTTP_NOTSUPP
     c                                     : ' using old format of '
     c                                     + 'http_close')
     c                   return    -1
     c                   endif

     c                   if        (wkDebugNoClose)
     c                   eval      wkDebugNoClose = *OFF                        Closed in HTTP
     c                   else
     c                   callp     debug_close
     c                   endif

     c                   if        comm_hangup ( peComm )
     c                   return    0
     c                   else
     c                   return    -1
     c                   endif
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  HTTP GET command:   This performs the logic of an HTTP GET.
      *
      *
      *          peSaveProc = procedure to use for saving received data
      *              peSock = pointer to socket descriptor/handle/structure
      *              peFile = file descriptor/handle/structure
      *           peTimeout = timeout in seconds.  If no data could be
      *                  sent or received in this time period, procedure
      *                  will return 0.
      *
      *           peAbsPath = absolute path to document on host.
      *              peHost = host to receive document from
      *           peModTime = (optional) Only download document if it's
      *                  modification time is newer than this timestamp.
      *                  If omitted, the document will always be retrieved
      *         peUserAgent = (optional) User-Agent string to report to
      *                  server.  If blank, no string is sent.  If omitted,
      *                  'http-api/1.0' will be sent.
      *        peConentType = (optional) Content-type reported to CGI
      *                  scripts.  If omitted or blank, none will be sent.
      *
      *   returns:     HTTP response code.
      *             or 0 upon timeout
      *             or -1 upon internal error (check HTTP_ERROR)
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P do_get          B
     D do_get          PI            10I 0
     D  peSaveProc                     *   value procptr
     D  peComm                         *   value
     D  peFile                       10I 0 value
     D  peTimeout                    10I 0 value
     D  peAbsPath                 32767A   varying const options(*varsize)
     D  peHost                      256A   const
     D  peModTime                      Z   options(*omit)
     D  peUserAgent                  64A   options(*omit)
     D  peContentType                64A   options(*omit)
     D  peSoapAction                 64A   options(*omit)
     D  pePort                       10I 0 value
     D  peSecure                      1N   value
     D  peServ                       32A   const

     D wwReqChain      S          32767A   varying
     D wwAddlHdr       s           1024A   varying
     D wwRespChain     s          32767A   varying
     D wwModString     S             29A
     D rc              S             10I 0
     D wwTE            S             32A
     D wwAuthErr       S              1N   inz(*OFF)
     D wwProxyAuthErr...
     D                 S              1N   inz(*OFF)
     D wwClen          s             10U 0
     D wwUseCL         s              1N
     D wwFinRC         S             10I 0
     D wwErrorNo       s             10I 0
     D wwErrorMsg      s                   like(http_error)
     d wwPathPfx       s           1024A   varying

     c                   callp     debug_msg('do_get(): entered')

     C*********************************************************
     C*  Build an HTTP/1.1 request chain:
     C*********************************************************
     c
     c* If Connection is done via Proxy, an URI instead of just the path has to be sent.
     c                   if        dsProxyHost <> *blanks
     c                               and dsProxyTun = *Off
     c                   eval      wwPathPfx = %trim(peServ) + '://' +
     c                             %trim(peHost)
     c                   if        pePort <> 0
     c                   eval      wwPathPfx = wwPathPfx + ':' +
     c                                        %trim(%editc(pePort:'L'))
     c                   endif
     c                   else
     c                   eval      wwPathPfx = ''
     c                   endif
     c
     c                   eval      wwReqChain =
     c                             'GET ' + wwPathPfx+%trim(peAbsPath) +
     c                                    ' HTTP/1.1'+CRLF+
     c                             'Host: '+%trim(peHost)

     c                   if        pePort = 0
     c                   eval      wwReqChain = wwReqChain + CRLF
     c                   else
     c                   eval      wwReqChain = wwReqChain + ':'
     c                                        + %trim(%editc(pePort:'L'))
     c                                        + CRLF
     c                   endif

     c                   if        dsProxyAuthType = HTTP_AUTH_BASIC
     c                                and dsProxyTun = *off
     c                   eval      wwReqChain = wwReqChain +
     c                             'Proxy-Authorization: Basic '+
     c                             dsProxyAuthStr + CRLF
     c                   endif

     c                   select
     c                   when      dsAuthType = HTTP_AUTH_BASIC
     c                   eval      wwReqChain = wwReqChain +
     c                             'Authorization: Basic ' +
     c                              dsAuthStr + CRLF
     c                   when      dsAuthType = HTTP_AUTH_MD5_DIGEST
     c                   eval      wwReqChain = wwReqChain +
     c                             'Authorization: Digest ' +
     c                             mkdigest('GET':peAbsPath) + CRLF
     c                   endsl

     c                   if        %addr(peModTime) <> *NULL
     c                   eval      wwModString = httpdate(peModTime)
     c                   eval      wwReqChain = wwReqChain +
     c                             'If-Modified-Since: '+wwModString+CRLF
     c                   endif

     c                   if        %addr(peUserAgent)<>*NULL
     c                   if        peUserAgent <> *blanks
     c                   eval      wwReqChain = wwReqChain +
     c                             'User-Agent: '+%trimr(peUserAgent)+CRLF
     c                   endif
     c                   else
     c                   eval      wwReqChain = wwReqChain +
     c                             'User-Agent: ' + HTTP_USERAGENT + CRLF
     c                   endif

     c                   if        %addr(peContentType) <> *NULL
     c                   if        peContentType <> *blanks
     c                   eval      wwReqChain = wwReqChain +
     c                             'Content-Type: '+%trimr(peContentType)+
     c                             CRLF
     c                   endif
     c                   endif

     c                   if        %addr(peSOAPAction) <> *NULL
     c                   if        peSOAPAction <> *blanks
     c                   eval      wwReqChain = wwReqChain +
     c                             'SOAPAction: ' +%trimr(peSOAPAction)+
     c                             CRLF
     c                   else
     c                   eval      wwReqChain = wwReqChain +
     c                             'SOAPAction:  ' + CRLF
     c                   endif
     c                   endif

     c                   if        use_cookies = *ON
     c                   eval      wwReqChain = wwReqChain +
     c                             header_get_req_cookies( %trim(peHost)
     c                                                   : %trim(peAbsPath)
     c                                                   : peSecure )
     c                   endif

     c                   if        wkAddHdrProc <> *NULL
     c                   callp     addl_headers(wwAddlHdr: wkAddHdrData)
     c                   eval      wwReqChain = wwReqChain + wwAddlHdr
     c                   endif

     c                   eval      wwReqChain = wwReqChain + CRLF
     C*********************************************************
     C*  Send request chain & receive the response chain
     C*********************************************************
     c                   eval      rc = SendReq( peComm
     c                                         : %addr(wwReqChain)+2
     c                                         : %len(wwReqChain)
     c                                         : peTimeout )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

     c                   eval      rc = RecvResp( peComm
     c                                          : wwRespChain
     c                                          : %size(wwRespChain)
     c                                          : peTimeout
     c                                          : *OFF )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

     C*********************************************************
     C*  Interpret the response chain:
     C*********************************************************
     c                   eval      rc = parse_resp_chain( wwRespChain
     c                                                  : rc
     c                                                  : wwTE
     c                                                  : wwCLen
     c                                                  : wwUseCL
     c                                                  : wwAuthErr
     c                                                  : wwProxyAuthErr
     c                                                  : %trim(peHost)
     c                                                  : %trim(peAbsPath))

     c                   if        rc<=100 or rc=204 or rc=304
     c                   return    rc
     c                   endif

     c                   eval      wwFinRC = rc
     c                   eval      wwErrorMsg = http_error(wwErrorNo)

     C*********************************************************
     C* receive the document
     C*********************************************************
     c                   if        %scan('chunked': wwTE) > 0
     c                   eval      rc = RecvChunk(peComm: peSaveProc:
     c                                    peFile: peTimeout)
     c                   else
     c                   eval      rc = RecvDoc(peComm: peSaveProc:
     c                                    peFile: peTimeout: wwCLen:
     c                                    wwUseCL)
     c                   endif

     c                   if        rc<1  or wwFinRC=200
     c                   return    rc
     c                   else
     c                   callp     SetError(wwErrorNo: wwErrorMsg)
     c                   return    wwFinRC
     c                   endif
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  HTTP POST command:   This performs the logic of an HTTP POST.
      *
      *          peSaveProc = procedure to use for saving received data
      *          peSendProc = procedure to use for sending a doc body.
      *
      *          pePostData = data to send to receiving CGI script
      *                (can be a pointer to a string or a user space, or
      *                randomly allocated memory.)
      *       pePostDataLen = length of post data
      *
      *              peSock = pointer to socket descriptor/handle/structure
      *              peFile = file descriptor/handle/structure
      *
      *           peTimeout = timeout in seconds.  If no data could be
      *                  sent or received in this time period, procedure
      *                  will return 0.
      *
      *           peAbsPath = absolute path to document on host.
      *              peHost = host to receive document from
      *         peUserAgent = (optional) User-Agent string to report to
      *                  server.  If blank, no string is sent.  If omitted,
      *                  'http-api/1.0' will be sent.
      *       peContentType = (optional) Content-type reported to CGI
      *                  scripts.  If omitted or blank, none will be sent.
      *
      *   returns:     HTTP response code.
      *             or 0 upon timeout
      *             or -1 upon internal error (check HTTP_ERROR)
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P do_post         B
     D do_post         PI            10I 0
     D  peSaveProc                     *   value procptr
     D  peSendProc                     *   value procptr
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peComm                         *   value
     D  peFile                       10I 0 value
     D  peTimeout                    10I 0 value
     D  peAbsPath                 32767A   const varying options(*varsize)
     D  peHost                      256A   const
     D  peUserAgent                  64A   options(*omit)
     D  peContentType                64A   options(*omit)
     D  peSoapAction                 64A   options(*omit)
     D  pePostProc                     *   value procptr
     D  pePostFD                     10I 0 value
     D  pePort                       10I 0 value
     D  peSecure                      1N   value
     D  peServ                       32A   const

     D SendProc        PR            10I 0 extproc(peSendProc)
     D  peComm                         *   value
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peTimeout                    10I 0 value
     D  pePostProc                     *   procptr value
     D  pePostFD                     10I 0 value

     D wwReqChain      S          32767A   varying
     D wwRespChain     S          32767A   varying
     D wwAddlHdr       s           1024A   varying
     D wwModString     S             29A
     D rc              S             10I 0
     D wwPos           S             10I 0
     D wwPos2          S             10I 0
     D wwTE            S             32A
     D wwAuthErr       S              1N   inz(*OFF)
     D wwProxyAuthErr...
     D                 S              1N   inz(*OFF)
     D wwCLen          S             10U 0
     D wwCL            S             32A
     D wwUseCL         s              1N
     D wwFinRC         S             10I 0
     D wwErrorNo       s             10I 0
     D wwErrorMsg      s                   like(http_error)
     D wwErr           s             10I 0
     d wwPathPfx       s           1024A   varying

     c                   callp     debug_msg('do_post(): entered')

     C*********************************************************
     C*  Build an HTTP/1.1 request chain:
     C*********************************************************
     c* If Connection is done via Proxy, an URI instead of just the path has to be sent.
     c                   if        dsProxyHost <> *blanks
     c                               and dsProxyTun = *Off
     c                   eval      wwPathPfx = %trim(peServ) + '://' +
     c                             %trim(peHost)
     c                   if        pePort <> 0
     c                   eval      wwPathPfx = wwPathPfx + ':' +
     c                                        %trim(%editc(pePort:'L'))
     c                   endif
     c                   else
     c                   eval      wwPathPfx = ''
     c                   endif
     c
     c                   eval      wwReqChain =
     c                             'POST ' + wwPathPfx+%trim(peAbsPath) +
     c                                    ' HTTP/1.1'+CRLF+
     c                             'Host: '+%trim(peHost)

     c                   if        pePort = 0
     c                   eval      wwReqChain = wwReqChain + CRLF
     c                   else
     c                   eval      wwReqChain = wwReqChain + ':'
     c                                        + %trim(%editc(pePort:'L'))
     c                                        + CRLF
     c                   endif

     c                   if        %addr(peUserAgent)<>*NULL
     c                   if        peUserAgent <> *blanks
     c                   eval      wwReqChain = wwReqChain +
     c                             'User-Agent: '+%trimr(peUserAgent)+CRLF
     c                   endif
     c                   else
     c                   eval      wwReqChain = wwReqChain +
     c                             'User-Agent: ' + HTTP_USERAGENT + CRLF
     c                   endif

     c                   if        %addr(peContentType) <> *NULL
     c                   if        peContentType <> *blanks
     c                   eval      wwReqChain = wwReqChain +
     c                             'Content-Type: '+%trimr(peContentType)+
     c                             CRLF
     c                   endif
     c                   else
     c                   eval      wwReqChain = wwReqChain +
     c                             'Content-Type: ' + HTTP_CONTTYPE + CRLF
     c                   endif

     c                   if        %addr(peSOAPAction) <> *NULL
     c                   if        peSOAPAction <> *blanks
     c                   eval      wwReqChain = wwReqChain +
     c                             'SOAPAction: ' +%trimr(peSOAPAction)+
     c                             CRLF
     c                   else
     c                   eval      wwReqChain = wwReqChain +
     c                             'SOAPAction:  ' + CRLF
     c                   endif
     c                   endif

     c                   eval      wwReqChain = wwReqChain +
     c                             'Expect: 100-continue' + CRLF
     c                   eval      wwReqChain = wwReqChain +
     c                             'Content-Length: ' +
     c                             %trim(%editc(pePostDataLen:'P')) + CRLF

     c                   if        dsProxyAuthType = HTTP_AUTH_BASIC
     c                                and dsProxyTun = *off
     c                   eval      wwReqChain = wwReqChain +
     c                             'Proxy-Authorization: Basic '+
     c                             dsProxyAuthStr + CRLF
     c                   endif

     c                   select
     c                   when      dsAuthType = HTTP_AUTH_BASIC
     c                   eval      wwReqChain = wwReqChain +
     c                             'Authorization: Basic ' +
     c                              dsAuthStr + CRLF
     c                   when      dsAuthType = HTTP_AUTH_MD5_DIGEST
     c                   eval      wwReqChain = wwReqChain +
     c                             'Authorization: Digest ' +
     c                              mkdigest('POST':peAbsPath) + CRLF
     c                   endsl

     c                   if        use_cookies = *on
     c                   eval      wwReqChain = wwReqChain +
     c                             header_get_req_cookies( %trim(peHost)
     c                                                   : peAbsPath
     c                                                   : peSecure )
     c                   endif

     c                   if        wkAddHdrProc <> *NULL
     c                   callp     addl_headers(wwAddlHdr: wkAddHdrData)
     c                   eval      wwReqChain = wwReqChain + wwAddlHdr
     c                   endif

     c                   eval      wwReqChain = wwReqChain + CRLF

      *********************************************************
      *  Send request chain & get response
      *********************************************************
     c                   eval      rc = SendReq( peComm
     c                                         : %addr(wwReqChain)+2
     c                                         : %len(wwReqChain)
     c                                         : peTimeout )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

     c                   eval      rc = RecvResp( peComm
     c                                          : wwRespChain
     c                                          : %size(wwRespChain)
     c                                          : wk100time
     c                                          : *ON )
     c                   if        rc<0
     c                   return    rc
     c                   endif

      *********************************************************
      * Interpret response.  It can be "100" meaning that
      * we continue with a POST request.  Or an error.
      *********************************************************
     c                   if        rc = 0

     c                   eval      wwFinRC = 100
     c                   eval      wwErrorMsg = 'CONTINUE'

     c                   else

     c                   eval      rc = parse_resp_chain( wwRespChain
     c                                                  : rc
     c                                                  : wwTE
     c                                                  : wwCLen
     c                                                  : wwUseCL
     c                                                  : wwAuthErr
     c                                                  : wwProxyAuthErr
     c                                                  : %trim(peHost)
     c                                                  : peAbsPath)
     c                   if        rc<100 or rc=204 or rc=304
     c                   return    rc
     c                   endif

     c                   eval      wwFinRC = rc
     c                   eval      wwErrorMsg = http_error(wwErrorNo)

     c                   endif

      *********************************************************
      * If we received the go-ahead, send the document
      *********************************************************
     c                   if        wwFinRC = 100

     c                   eval      rc = SendProc( peComm
     c                                          : pePostData
     c                                          : pePostDataLen
     c                                          : peTimeout
     c                                          : pePostProc
     c                                          : pePostFD )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

     c                   eval      rc = RecvResp( peComm
     c                                          : wwRespChain
     c                                          : %size(wwRespChain)
     c                                          : peTimeout
     c                                          : *Off )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

     c                   eval      rc = parse_resp_chain( wwRespChain
     c                                                  : rc
     c                                                  : wwTE
     c                                                  : wwCLen
     c                                                  : wwUseCL
     c                                                  : wwAuthErr
     c                                                  : wwProxyAuthErr
     c                                                  : %trim(peHost)
     c                                                  : peAbsPath )

     c                   if        rc<100 or rc=204 or rc=304
     c                   return    rc
     c                   endif

     c                   eval      wwFinRC = rc
     c                   eval      wwErrorMsg = http_error(wwErrorNo)

     c                   endif

     C*********************************************************
     C* receive the document from the server
     C*********************************************************
     c                   if        %scan('chunked': wwTE) > 0
     c                   eval      rc = RecvChunk(peComm: peSaveProc:
     c                                    peFile: peTimeout)
     c                   else
     c                   eval      rc = RecvDoc(peComm: peSaveProc:
     c                                    peFile: peTimeout: wwCLen:
     c                                    wwUseCL)
     c                   endif

     c                   if        rc<1 or wwFinRC=200
     c                   return    rc
     c                   else
     c                   callp     SetError(wwErrorNo: wwErrorMsg)
     c                   return    wwFinRC
     c                   endif
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * SendReq():  Send request chain
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P SendReq         B
     D SendReq         PI            10I 0
     D   peComm                        *   value
     D   peData                        *   value
     D   peDataLen                   10I 0 value
     D   peTimeout                   10I 0 value
     D p_deref         s               *
     D wwDeref         s              1A   based(p_deref)

     c                   callp     http_xlatep( peDataLen
     c                                        : peData
     c                                        : TO_ASCII )

     c                   return    comm_BlockWrite( peComm
     c                                            : peData
     c                                            : peDataLen
     c                                            : peTimeout )
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  recvresp():  Receives an HTTP response chain from the server.
      *
      *      peComm = Comm driver to use when receiving
      * peRespChain = complete request chain to sent
      *   peRespLen = length of request chain data
      *  peTimeOut  = Timeout value in seconds.  If no data can be sent
      *         for this amount of time, it will return a timeout.
      *    peUse100 = Return from recvresp() if a 100-continue comes up
      *
      *  Returns 1 upon success, 0 upon timeout, -1 upon error
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P recvresp        B
     D recvresp        PI            10I 0
     D   peComm                        *   value
     D   peRespChain              32767A   varying options(*varsize)
     D   peRespLen                   10I 0 value
     D   peTimeOut                   10P 3 value
     D   peUse100                     1N   const

     D wwPos           S               *
     D wwRec           S             10I 0
     D wwLen           S             10I 0
     D p_saveaddr      S               *
     D p_check         S               *
     D wwCheck         S              4A   based(p_check)
     D wwRespCode      S             10I 0
     D forever         S              1N   inz(*On)
     D repeating       S              1N   inz(*Off)
     D wwSecs          s             10I 0
     D wwMicroSecs     s             10I 0
     D wwLeft          s             10I 0
     D wwErr           s             10I 0
     D CR              S              1A   inz(x'0d') static

     c                   callp     debug_msg('recvresp(): entered')

     c                   dou       not repeating

     c                   eval      wwPos = %addr(peRespChain) + 2
     c                   eval      wwLeft = peRespLen - 2
     c                   eval      %len(peRespChain) = wwLeft
     c                   eval      wwLen = 0

     c                   dow       wwLeft > 0

     c                   eval      wwRec = comm_lineread( peComm
     c                                                  : wwPos
     c                                                  : wwLeft
     c                                                  : peTimeout )

     c                   if        wwRec < 1
     c                   callp     http_error(wwErr)
     c                   if        wwErr = HTTP_BRTIME
     c                   return    0
     c                   else
     c                   return    -1
     c                   endif
     c                   endif

     c                   eval      wwPos = wwPos + wwRec
     c                   eval      wwLeft = wwLeft - wwRec
     c                   eval      wwLen = wwLen + wwRec

     c                   if        wwLen >= 4
     c                   eval      p_check = wwPos - 4
     c                   if        wwCheck = x'0d0a0d0a'
     c                   leave
     c                   endif
     c                   endif

     c                   enddo

     c                   eval      %len(peRespChain) = wwLen

     C* translate response chain to EBCDIC
     c                   if        wwLen > 0
     c                   callp     http_xlatep( wwLen
     c                                        : %addr(peRespChain) + 2
     c                                        : TO_EBCDIC )
     c                   endif

     C* check for "continue" type codes:
     C* if we get them, we'll look for a whole new chain :)
     c                   if        %subst(peRespChain:1:2) = x'0d25'
     c                   eval      p_check = %addr(peRespChain) + 13
     c                   else
     c                   eval      p_check = %addr(peRespChain) + 11
     c                   endif
     c                   eval      wwRespCode = atoi(wwCheck)

     c                   select
     c                   when      peUse100 and wwRespCode = 100
      *                 we are handling this specially in do_post
     c                   eval      repeating = *off
     c                   when      wwRespCode > 99 and wwRespCode < 200
     c                   eval      repeating = *on
     c                   other
     c                   eval      repeating = *off
     c                   endsl

     c                   if        %scan(CR:peRespChain) > 1
     c                   callp     SetError(HTTP_RESP: %subst(peRespChain:1:
     c                                 %scan(CR:peRespChain)-1) )
     c                   else
     c                   callp     SetError(HTTP_RESP: peRespChain)
     c                   endif

     c                   enddo

     c                   return    wwRespCode
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * recvdoc(): receive (Download) http document
      *
      *       peComm = Comm driver to receive with
      *  peProcedure = procedure to call with received data.
      *    peTimeout = time-out in seconds.  If no data is received
      *          for this amount of time, proc will time out.
      *
      *  returns 1 upon success, 0 upon timeout, -1 upon failure
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P recvdoc         B
     D recvdoc         PI            10I 0
     D   peComm                        *   value
     D   peProcedure                   *   value procptr
     D   peFD                        10I 0 value
     D   peTimeout                   10I 0 value
     D   peCLen                      10U 0 value
     D   peUseCL                      1N   const

     D saveproc        PR            10I 0 ExtProc(peProcedure)
     D  fd                           10I 0 value
     D  data                           *   value
     D  length                       10I 0 value

     D wwData          S           8192A
     D forever         S              1N   inz(*on)
     D wwErr           S             10I 0
     D p_saveaddr      S               *
     D wwTimeout       S              8A
     D wwLen           S             10I 0
     D rc              S             10I 0
     D wwChunked       S              1N
     D wwChunk         S             10I 0
     D wwReceived      S             10U 0
     D wwRet           s             10I 0

     c                   callp     debug_msg('recvdoc(): entered')
     c                   callp     SetError(0: *blanks)
     c                   eval      wwReceived = 0

     c                   if        peUseCL = *OFF
     c                   eval      wwRet = 1
     c                   else
     c                   eval      wwRet = -1
     c                   endif

     c                   if        peUseCL=*ON and peCLen=0
     c                   return    1
     c                   endif

     c                   dow       forever

     c                   eval      wwLen = comm_read( peComm
     c                                              : %addr(wwData)
     c                                              : %size(wwData)
     c                                              : peTimeout     )

     c                   if        wwLen < 0
     c                   return    wwRet
     c                   endif

     c                   eval      rc = saveproc( peFD
     c                                          : %addr(wwData)
     c                                          : wwLen )
     c                   if        rc < wwLen
     c                   callp     SetError(HTTP_RDWERR
     c                                     : 'errno is currently '
     c                                     + %trim(%editc(errno:'L')))
     c                   callp     SetError(HTTP_RDWERR:'recvdoc: saveproc:'+
     c                              ' Not all data was written!')
     c                   return    -1
     c                   endif

     c                   eval      wwReceived = wwReceived + wwLen
     c                   if        peUseCL=*ON and wwReceived>=peCLen
     c                   return    1
     c                   endif

     C* Call status proc if defined
     c                   if        wkDwnlProc <> *NULL
     c                   callp     download_sts(wwReceived: peCLen
     c                                         : wkDwnlUdata)
     c                   endif

     c                   enddo

     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * recvchunk(): receive (download) data using chunked transfer-encoding
      *
      *       peComm = Comm driver to receive from
      *  peProcedure = procedure to call with received data.
      *    peTimeout = time-out in seconds.  If no data is received
      *          for this amount of time, proc will time out.
      *
      *  returns 1 upon success, 0 upon timeout, -1 upon failure
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P recvchunk       B
     D recvchunk       PI            10I 0
     D   peComm                        *   value
     D   peProcedure                   *   value procptr
     D   peFD                        10I 0 value
     D   peTimeout                   10I 0 value

     D saveproc        PR            10I 0 ExtProc(peProcedure)
     D  fd                           10I 0 value
     D  data                           *   value
     D  length                       10I 0 value

     D wwData          s           8192A
     D forever         S              1N   inz(*on)
     D rc              S             10I 0
     D wwLeft          S             10I 0
     D wwRecSize       S             10I 0
     D wwCRLF          s              2A
     D wwReceived      S             10U 0

     c                   callp     debug_msg('recvchunk(): entered')
     c                   eval      wwReceived = 0

     c                   dow       Forever

     C*********************************************************
     C* Receive the size of the next chunk of data:
     C*********************************************************
     c                   eval      wwLeft = get_chunk_size(peComm:peTimeout)
     c                   callp     debug_msg('get_chunk_size returned ' +
     c                                          %trim(%editc(wwLeft:'P')))
     c                   select
     c                   when      wwLeft = 0
     c                   return    1
     c                   when      wwLeft = -2
     c                   return    0
     c                   when      wwLeft = -1
     c                   return    -1
     c                   endsl

     C*********************************************************
     c* Receive data until we have an entire chunk:
     C*********************************************************
     c                   dou       wwLeft = 0

     c                   eval      wwRecSize = %size(wwData)
     c                   if        wwLeft < wwRecSize
     c                   eval      wwRecSize = wwLeft
     c                   endif

     c                   callp     debug_msg('calling comm_blockread')

     c                   if        comm_BlockRead( peComm
     c                                           : %addr(wwData)
     C                                           : wwRecSize
     c                                           : peTimeout ) < wwRecSize
     c                   callp     debug_msg('comm_blockread failed!')
     c                   return    -1
     c                   endif

     c                   callp     debug_msg('comm_blockread returned ' +
     c                              %trim(%editc(wwRecSize:'P')) )

     C* Write any received data to the save procedure:
     c                   eval      rc = saveproc(peFD: %addr(wwData):
     c                                            wwRecSize)
     c                   if        rc < wwRecSize
     c                   callp     SetError(HTTP_RDWERR:'recvchunk: saveproc:'+
     c                              ' Not all data was written!')
     c                   return    -1
     c                   endif

     c                   if        wkDwnlProc <> *NULL
     c                   eval      wwReceived = wwReceived + wwRecSize
     c                   callp     download_sts(wwReceived: 0: wkDwnlUData)
     c                   endif

     c                   eval      wwLeft = wwLeft - wwRecSize
     c                   enddo

     C*********************************************************
     c* Receive the CRLF that follows each chunk
     C*********************************************************
     c                   if        comm_BlockRead( peComm
     c                                           : %addr(wwCRLF)
     c                                           : %size(wwCRLF)
     c                                           : peTimeout ) < 2
     c                             or wwCRLF <> x'0d0a'
     c                   callp     SetError(HTTP_RDCRLF: 'recvchunk: '+
     c                                'No CRLF after reading chunk!')
     c                   return    -1
     c                   endif

     c                   enddo
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  This sends a document body, such as those used by the PUT
      *  or POST HTTP commands.
      *
      *  returns 0 for timeout, -1 for error or 1 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P SendDoc         B
     D SendDoc         PI            10I 0
     D  peComm                         *   value
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peTimeout                    10I 0 value
     D  peUnused1                      *   value procptr
     D  peUnused2                    10I 0 value

     D wwPos           S               *
     D wwLeft          S             10I 0
     D wwTimeout       S              8A
     D wwErr           S             10I 0
     D wwSent          S             10I 0
     D wwChunk         S             10I 0

     c                   callp     debug_msg('senddoc(): entered')

     c                   eval      wwPos = pePostData
     c                   eval      wwLeft = pePostDataLen

     c                   dow       wwLeft > 0

     c                   eval      wwChunk = wwLeft
     c                   if        wwChunk > 8192
     c                   eval      wwChunk = 8192
     c                   endif

     c                   eval      wwSent = comm_BlockWrite( peComm
     c                                                     : wwPos
     c                                                     : wwChunk
     c                                                     : peTimeout )

     c                   eval      wwLeft = wwLeft - wwSent
     c                   eval      wwPos  = wwPos + wwSent

     c                   if        wkUplProc <> *NULL
     c                   callp     upload_sts(pePostDataLen-wwLeft
     c                                       : pePostDataLen
     c                                       : wkUplUData )
     c                   endif

     c                   enddo

     c                   return    1
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  This sends a document body, such as those used by the PUT
      *  or POST HTTP commands, but uses a callback instead of a buffer
      *
      *  returns 0 for timeout, -1 for error or 1 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P SendRaw         B
     D SendRaw         PI            10I 0
     D  peComm                         *   value
     D  peUnused1                      *   value
     D  peDataSize                   10I 0 value
     D  peTimeout                    10I 0 value
     D  pePostProc                     *   value procptr
     D  pePostFD                     10I 0 value

     D Callback        PR            10I 0 extproc(pePostProc)
     D   fd                          10I 0 value
     D   data                      8192A
     D   size                        10I 0 value

     D wwLen           S             10I 0
     D wwBuf           s           8192A
     D wwSent          S             10I 0
     D wwNeed          S             10I 0

     c                   callp     debug_msg('sendraw(): entered')
     c                   eval      wwSent = 0

     c                   dow       peDataSize > wwSent

     c                   eval      wwNeed = %size(wwBuf)
     c                   if        wwNeed > peDataSize
     c                   eval      wwNeed = peDataSize
     c                   endif

     c                   eval      wwLen = Callback( pePostFD
     c                                             : wwBuf
     c                                             : wwNeed )

     c                   if        wwLen < 1
     c                   callp     SetError( HTTP_SWCERR
     c                                     : 'SendRaw(): callback '
     c                                     + 'returned an error.')
     c                   return    -1
     c                   endif

     c                   if        wwLen > wwNeed
     c                   callp     SetError( HTTP_SWCERR
     c                                     : 'SendRaw(): callback '
     c                                     + 'supplied too much data.')
     c                   return    -1
     c                   endif

     c                   if        comm_BlockWrite( peComm
     c                                            : %addr(wwBuf)
     c                                            : wwLen
     c                                            : peTimeout
     c                                            ) < wwLen
     c                   return    -1
     c                   endif

     c                   eval      wwSent = wwSent + wwLen

     c                   if        wkUplProc <> *NULL
     c                   callp     upload_sts(wwSent
     c                                       : peDataSize
     c                                       : wkUplUData )
     c                   endif

     c                   enddo

     c                   return    wwSent
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  This receives the chunk size from the http stream.  We use
      *   it so we know how big the next chunk of data is.
      *
      *  XXX: This will crash if 2gb or larger chunks are used.
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P get_chunk_size  B
     d get_chunk_size  PI            10I 0
     d   peComm                        *   value
     d   peTimeout                   10I 0 value

     d sscanf          PR            10I 0 extproc('sscanf')
     D  str                            *   value options(*string)
     D  format                         *   value options(*string)
     D  sexyfun                      10U 0

     D wwLen           s             10I 0
     D rc              s             10I 0
     D wwBuf           s             15A
     D wwChunkSize     S             10U 0
     D wwErr           s             10I 0

     c                   callp     debug_msg('get_chunk_size(): entered')

     c                   eval      wwLen = comm_LineRead( peComm
     c                                                  : %addr(wwBuf)
     c                                                  : %size(wwBuf)
     c                                                  : peTimeout )

     c                   if        wwLen < 1
     c                   callp     http_error(wwErr)
     c                   if        wwErr = HTTP_BRTIME
     c                   return    -2
     c                   else
     c                   return    -1
     c                   endif
     c                   endif

     c                   callp     http_xlate( wwLen
     c                                       : wwBuf
     c                                       : TO_EBCDIC )

     c                   eval      rc = %scan(';': wwBuf)
     c                   if        rc > 1
     c                   eval      wwBuf = %subst(wwBuf:1:rc-1)
     c                   endif

     c                   callp     sscanf(wwBuf: '%x': wwChunkSize)

     c                   callp     debug_msg('chunk size = ' +
     c                                  %trim(%editc(wwChunkSize:'P')))

     c                   return    wwChunkSize
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_getauth():   Get HTTP Authentication Information
      *
      *   Call this proc after you receive a HTTP_NDAUTH error
      *   to determine the authentication credentials that are required
      *
      *  The following parms are returned to your program:
      *
      *     peBasic = *ON if BASIC auth is allowed
      *    peDigest = *ON if MD5 DIGEST auth is allowed
      *     peRealm = Auth realm.  Present this to the user to identify
      *               which password you're looking for.  For example
      *               if peRealm is "secureserver.com" you might say
      *               "enter password for secureserver.com" to user.
      *
      *   After getting the userid & password from the user (or database)
      *   you'll need to call http_setauth()
      *
      *  Returns -1 upon error, or 0 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_getauth    B                   export
     D http_getauth    PI            10I 0
     D   peBasic                      1N
     D   peDigest                     1N
     D   peRealm                    124A

     c                   callp     debug_msg('http_getauth(): entered')
     c                   if        dsAuthRealm = *blanks
     c                   callp     SetError(HTTP_NOAUTH: 'Server did ' +
     c                              'not ask for authentication!')
     c                   return    -1
     c                   endif

     c                   eval      peBasic = dsAuthBasic
     c                   eval      peDigest = dsAuthDigest
     c                   eval      peRealm = dsAuthRealm

     c                   return    0
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_setauth():   Set HTTP Authentication Information
      *
      *     peAuthType = Authentication Type (HTTP_AUTH_BASIC or
      *                     HTTP_AUTH_MD5_DIGEST)
      *     peUsername = UserName to use
      *     pePasswd   = Password to use
      *
      *  Returns -1 upon error, or 0 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_setauth    B                   export
     D http_setauth    PI            10I 0
     D   peAuthType                   1A   const
     D   peUsername                  80A   const
     D   pePasswd                  1024A   const

     D wwString        S           1105A
     D wwEncoded       S           1476A
     D wwEncLen        S             10I 0

     c                   callp     debug_msg('http_setauth(): entered')
     c                   if        peAuthType<>HTTP_AUTH_BASIC
     c                               and peAuthType<>HTTP_AUTH_MD5_DIGEST
     c                               and peAuthType<>HTTP_AUTH_NONE
     c                   callp     SetError(HTTP_ATHTYP: 'Invalid authenti'+
     c                                   'cation type!')
     c                   return    -1
     c                   endif

     C*************************************************
     c* Calculate strings for NO authentication
     C*************************************************
     c                   if        peAuthType = HTTP_AUTH_NONE
     c                   eval      dsAuthType = HTTP_AUTH_NONE
     c                   eval      dsAuthStr = ''
     c                   return    0
     c                   endif

     c                   if        %scan(':':peUserName) > 0
     c                              or %scan(':': pePasswd) > 0
     c                   callp     SetError(HTTP_ATHVAL: 'HTTP Auth value' +
     c                               's cannot contain a colon!')
     c                   return    -1
     c                   endif

     c                   eval      dsAuthUser = peUserName
     c                   eval      dsAuthPasswd = pePasswd

     c                   eval      wwString = %trimr(peUserName) + ':' +
     c                                        %trimr(pePasswd)
     c                   callp     http_xlate( %len(%trimr(wwString))
     c                                       : wwString
     c                                       : TO_ASCII )

     c                   eval      wwEncLen = base64_encode(%addr(wwString):
     c                                             %len(%trimr(wwString)):
     c                                             %addr(wwEncoded):
     c                                             %size(wwEncoded))

     c                   eval      dsAuthStr = %subst(wwEncoded:1:wwEncLen)
     c                   eval      dsAuthType = peAuthType

     c                   return    0
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * setUrlAuth(): Set auth credentials if found in the URL
      *      peUsername = username found in URL
      *      pePasswd   = password found in URL
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P setUrlAuth      B
     D setUrlAuth      PI
     D   peUsername                  80A   const
     D   pePasswd                  1024A   const
     c                   if        peUsername = *blanks
     c                              and pePasswd = *blanks
     c                   eval      wkSaveAuth = *blanks
     c                   return
     c                   endif

     c                   eval      wkSaveAuth = dsAuth

     c                   if        dsAuthRealm = *blanks
     c                               or dsAuthDigest = *OFF
     c                   callp     http_setAuth( HTTP_AUTH_BASIC
     c                                         : peUsername
     c                                         : pePasswd )
     c                   else
     c                   callp     http_setAuth( HTTP_AUTH_MD5_DIGEST
     c                                         : peUsername
     c                                         : pePasswd )
     c                   endif
     c                   return
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  mkdigest():  Create a digest authorization string
      *
      *      peMethod = HTTP method in use (GET or POST)
      *
      *  Returns the digest authorization response
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P mkdigest        B
     D mkdigest        PI         32767A   varying
     D   peMethod                     4A   const
     D   peURI                    32767A   varying const options(*varsize)

     D wwRet           S          32767A   varying
     D wwWork          S          32767A   varying
     D wwA1            S             32A
     D wwA2            S             32A
     D wwResp          S             32A

     c                   callp     debug_msg('mkdigest(): entered')
     C*  Authorization: Digest username="Mufasa",
     C*         realm="testrealm@host.com",
     C*         nonce="dcd98b7102dd2f0e8b11d0f600bfb0c093",
     C*         uri="/dir/index.html",
     C*         nc=00000001,
     C*         cnonce="0a4f113b",
     C*         response="6629fae49393a05397450978507c4ef1",
     C*         opaque="5ccc069c403ebaf9f0171e9517f40e41"

     c*  response =
     C*     md5(md5(A1) + ':' + nonce + ':' + nc-value + ':' +
     c*          cnonce + ':' + qop + ':' md5(A2))
     C*  A1 = username + ':' + realm + ':' + passwd
     C*  A2 = method + ':' + uri
     C*
     C*  realm = passed from server
     C*  nonce = passed from server
     C*  nc = count of uses of nonce so far
     C*  qop = auth (if that's a choice on this server)
     C*  cnonce = 8-char hex string that we make up
     C*  opaque = passed by server
     C*  method = GET or POST
     C*  username & password are supplied by the user
     C*  we should know the URI when the request is made

      * FIXME: This CNonce ("client nonce") should be different
      *    with every request, not a fixed string.
     c                   eval      dsAuthCNonce = '7248e2a3711545e8'
     c                   eval      dsAuthNC = dsAuthNC + 1

     c                   eval      wwWork = %trim(dsAuthUser) + ':' +
     c                                      %trim(dsAuthRealm) + ':' +
     c                                      %trim(dsAuthPasswd)
     c                   if        md5(%addr(wwWork)+2:%len(wwWork):wwA1)
     c                              = *OFF
     c                   return    '*error'
     c                   endif

     c                   eval      wwWork = %trim(peMethod) + ':' +
     c                                      %trim(peURI)
     c                   if        md5(%addr(wwWork)+2:%len(wwWork):wwA2)
     c                              = *OFF
     c                   return    '*error'
     c                   endif

     c                   if        %scan('auth': dsAuthQop) > 0
     c                   eval      wwWork = wwA1 + ':' +
     c                                      %trim(dsAuthNonce) + ':' +
     c                                      '0'+%editc(dsAuthNC:'X') + ':' +
     c                                      dsAuthCnonce + ':' +
     c                                      'auth' + ':' +
     c                                      wwA2
     c                   else
     c                   eval      wwWork = wwA1 + ':' +
     c                                      %trim(dsAuthNonce) + ':' +
     c                                      wwA2
     c                   endif

     c                   if        md5(%addr(wwWork)+2:%len(wwWork):wwResp)
     c                              = *OFF
     c                   return    '*error'
     c                   endif

     c                   eval      wwRet =
     c                               'username="' +%trim(dsAuthUser)+ '", '+
     c                               'realm="' +%trim(dsAuthRealm)+ '", '+
     c                               'nonce="' +%trim(dsAuthNonce)+ '", '+
     c                               'uri="' +%trim(peURI)+ '", '+
     c                               'response="' + wwResp + '"'

     c                   if        %scan('auth': dsAuthQop) > 0
     c                   eval      wwRet = %trimr(wwRet) + ', ' +
     c                               'algorithm=MD5, ' +
     c                               'nc=0' + %editc(dsAuthNC:'X') + ', ' +
     c                               'cnonce="' + dsAuthCNonce +'", ' +
     c                               'qop="auth"'
     c                   endif

     c                   if        dsAuthOpaque <> *blanks
     c                   eval      wwRet = %trimr(wwRet) + ', ' +
     c                                 'opaque="' + %trim(dsAuthOpaque)+'"'
     c                   endif

     c                   return    wwRet
     P                 E


     P*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P* Interpret (parse & save) the WWW-Authenticate: header
     P*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P interpret_auth  B
     D interpret_auth  PI
     D   peRespChain               2048A   const
     D   peKwdPos                    10I 0 value

     D wwCh            S              1A
     D wwAuth          S           2048A
     D wwPos1          S             10I 0
     D wwPos2          S             10I 0
     D wwLen           S             10I 0
     D p_Word          S               *
     D p_Next          S               *
     D wwWord          S             30A
     D TAB             C                   CONST(x'05')
     D wwOrigNonce     S            128A
     D LF              S              1A   inz(x'25') static

     d strtok          PR              *   extproc('strtok')
     d   nextpos                       *   value
     d   delim                         *   value options(*string)

     c                   callp     debug_msg('interpret_auth(): entered')
     c                   eval      wwOrigNonce = dsAuthNonce
     c                   eval      dsAuthBasic = *OFF
     c                   eval      dsAuthDigest = *OFF
     c                   eval      dsAuthDigest = *OFF
     c                   eval      dsAuthRealm = *blanks
     c                   eval      dsAuthNonce = *blanks
     c                   eval      dsAuthOpaque = *blanks
     c                   eval      dsAuthQOP = *blanks
     c                   eval      dsAuthCNonce = *blanks

     C*********************************************************
     C* extract the value of this keyword (and nothing else)
     C*********************************************************
     c                   eval      wwPos1 =%scan(':':peRespChain:peKwdPos+1)
     c                   eval      peKwdPos = wwPos1 + 1

     c                   dou       wwCh<>' ' and wwCh<>TAB
     c                   eval      wwPos2 = %scan(LF: peRespChain:
     c                                            wwPos1)
     c                   if        wwPos2 < wwPos1
     c                   leave
     c                   endif
     c                   eval      wwPos1 = wwPos2
     c                   eval      wwCh = %subst(peRespChain:wwPos2+1:1)
     c                   enddo

     c                   if        wwPos1 < peKwdPos
     C*             shouldn't happen
     c                   return
     c                   endif

     c                   eval      wwLen = (wwPos1 - peKwdPos) + 1
     c                   eval      wwAuth = %subst( peRespChain:
     c                                              peKwdPos: wwLen) + x'00'
     c                   eval      p_Next = %addr(wwAuth)

     C*********************************************************
     C* Extract one word at a time from the list.  If the word
     C* is something we can use, save it's value to our DS
     C*********************************************************
     c                   dow       1 = 1

     c                   eval      p_Word = strtok(p_Next: ' =,'+TAB)
     c                   if        p_Word = *NULL
     c                   leave
     c                   endif
     c                   eval      p_Next = *NULL

     c                   eval      wwWord = %str(p_Word)

     c     upper:lower   xlate     wwWord        wwWord

     c                   select
     c                   when      wwWord=*blanks
     c                   iter

     c                   when      wwWord = 'basic'
     c                   eval      dsAuthBasic = *on

     c                   when      wwWord = 'digest'
     c                   eval      dsAuthDigest = *on

     c                   when      wwWord = 'realm'
     c                   eval      p_Word = strtok(*NULL: '"')
     c                   if        p_word <> *NULL
     c                   eval      dsAuthRealm = %str(p_word)
     c                   endif

     c                   when      wwWord = 'qop'
     c                   eval      p_Word = strtok(*NULL: '"')
     c                   if        p_word <> *NULL
     c                   eval      dsAuthQOP = %str(p_word)
     c                   endif

     c                   when      wwWord = 'nonce'
     c                   eval      p_Word = strtok(*NULL: '"')
     c                   if        p_word <> *NULL
     c                   eval      dsAuthNonce = %str(p_word)
     c                   if        wwOrigNonce <> dsAuthNonce
     c                   eval      dsAuthNC = 0
     c                   endif
     c                   endif

     c                   when      wwWord = 'opaque'
     c                   eval      p_Word = strtok(*NULL: '"')
     c                   if        p_word <> *NULL
     c                   eval      dsAuthOpaque = %str(p_word)
     c                   endif
     c                   endsl

     c                   enddo

     c                   return
     P                 E

     P*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P* Interpret (parse & save) the Proxy-Authenticate: header
     P*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P interpret_proxy_auth...
     P                 B
     D interpret_proxy_auth...
     D                 PI
     D   peRespChain               2048A   const
     D   peKwdPos                    10I 0 value

     D wwCh            S              1A
     D wwAuth          S           2048A
     D wwPos1          S             10I 0
     D wwPos2          S             10I 0
     D wwLen           S             10I 0
     D TAB             C                   CONST(x'05')
     D LF              S              1A   inz(x'25') static

     c                   callp     debug_msg('interpret_proxy_auth(): entered')

     C*********************************************************
     C* extract the value of this keyword (and nothing else)
     C*********************************************************
     c                   eval      wwPos1 =%scan(':':peRespChain:peKwdPos+1)
     c                   eval      peKwdPos = wwPos1 + 1

     c                   dou       wwCh<>' ' and wwCh<>TAB
     c                   eval      wwPos2 = %scan(LF: peRespChain:
     c                                            wwPos1)
     c                   if        wwPos2 < wwPos1
     c                   leave
     c                   endif
     c                   eval      wwPos1 = wwPos2
     c                   eval      wwCh = %subst(peRespChain:wwPos2+1:1)
     c                   enddo

     c                   if        wwPos1 < peKwdPos
     C*             shouldn't happen
     c                   return
     c                   endif

     c                   eval      wwLen = (wwPos1 - peKwdPos) + 1
     c                   eval      wwAuth = %subst( peRespChain:
     c                                              peKwdPos: wwLen)

     C*********************************************************
     C* Check if we deal with basic authentication
     C*********************************************************
     c     upper:lower   xlate     peRespChain   wwAuth

     c                   eval      wwPos1 = %scan('basic':wwAuth)
     c                   if        wwPos1 > 0
     c                   eval      dsProxyAuthBasic = *on

     c* Extract the realm, if provided
     c                   eval      wwPos1 = %scan('realm="': wwAuth:
     c                                  wwPos1)
     c                   if        wwPos1 > 0
     c                   eval      wwPos1 = wwPos1 + 7
     c                   eval      wwPos2 = %scan('"': wwAuth: wwPos1)
     c                   if        wwPos2 > wwPos1
     c                   eval      wwLen = wwPos2 - wwPos1
     c                   eval      dsProxyAuthRealm = %subst(peRespChain:
     c                                  wwPos1: wwLen)
     c                   endif
     c                   endif
     c                   endif

     c                   return
     P                 E

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_xproc():  Register a procedure to be called back at
      *                 a given exit point
      *
      *     peExitPoint = exit point.  Should be one of the constants
      *                HTTP_POINT_XXX defined in the HTTPAPI_H member
      *          peProc = address of procedure to call for this
      *                exit point. (pass *NULL to disable this point)
      *      peUserData = Pointer to user data. This will be passed
      *                to your call-back procedure. You can set it to
      *                *NULL if you don't need/want it.
      *
      *  Returns -1 upon error, or 0 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_xproc      B                   export
     D http_xproc      PI            10I 0
     D  peExitPoint                  10I 0 value
     D  peProc                         *   procptr value
     D  peUserData                     *   value options(*nopass)

     D wwUserData      s               *   inz(*NULL)

     c                   if        %parms >= 3
     c                   eval      wwUserData = peUserData
     c                   endif

     c                   select
     c                   when      peExitPoint = HTTP_POINT_DEBUG
     c                   callp     debug_setproc(peProc: wwUserData)
     c                   when      peExitPoint = HTTP_POINT_UPLOAD_STATUS
     c                   eval      wkUplProc = peProc
     c                   eval      wkUplUData = wwUserData
     c                   when      peExitPoint = HTTP_POINT_DOWNLOAD_STATUS
     c                   eval      wkDwnlProc = peProc
     c                   eval      wkDwnlUData = wwUserData
     c                   when      peExitPoint = HTTP_POINT_ADDL_HEADER
     c                   eval      wkAddHdrProc = peProc
     c                   eval      wkAddHdrData = wwUserData
     c                   when      peExitPoint = HTTP_POINT_PARSE_HEADER
     c                   eval      wkParseHdrProc = peProc
     c                   eval      wkParseHdrData = wwUserData
     c                   when      peExitPoint = HTTP_POINT_PARSE_HDR_LONG
     c                   eval      wkParseHdrLong = peProc
     c                   eval      wkParseHdrLongData = wwUserData
     c                   when      peExitPoint = HTTP_POINT_PARSE_HDR_LONG
     c                   eval      wkParseHdrLong = peProc
     c                   eval      wkParseHdrLongData = wwUserData
      /if defined(HAVE_SSLAPI)
     c                   when      peExitPoint = HTTP_POINT_CERT_VAL
     c                   callp     commssl_setxproc( peExitPoint
     c                                             : peProc
     c                                             : wwUserData )
     c                   when      peExitPoint = HTTP_POINT_GSKIT_CERT_VAL
     c                   callp     commssl_setxproc( peExitPoint
     c                                             : peProc
     c                                             : wwUserData )
      /endif
     c                   other
     c                   callp     SetError(HTTP_BADPNT: 'Invalid exit ' +
     c                               'point!')
     c                   return    -1
     c                   endsl

     c                   return    0
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * http_redir_loc(): Retrieve location provided by a redirect
      *   request.
      *
      *  returns redirect location, or '' if no redirect was given
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_redir_loc  B                   export
     D http_redir_loc  PI          1024A   varying
     c                   return    wkRedirLoc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_long_ParseURL(): Parse URL into it's component parts
      *
      *  Breaks a uniform resource locator (URL) into it's component
      *  pieces for use with the http: or https: protocols.  (would also
      *  work for FTP with minor tweaks)
      *
      *  peURL = URL that needs to be parsed.
      *  peService = service name from URL (i.e. http or https)
      *  peUserName = user name given, or *blanks
      *  pePassword = password given, or *blanks
      *  peHost = hostname given in URL. (could be domain name or IP)
      *  pePort = port number to connect to, if specified, otherwise 0.
      *  pePath = remaining path/request for server.
      *
      *  returns -1 upon failure, or 0 upon success
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_long_ParseURL...
     P                 B                   export
     d http_long_ParseURL...
     D                 PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  peService                    32A
     D  peUserName                   32A
     D  pePassword                   32A
     D  peHost                      256A
     D  pePort                       10I 0
     D  pePath                    32767A   varying

     D wwLen           S             10I 0
     D wwURL           S          32767A   varying
     D wwTemp          S             65A
     D wwPos           S             10I 0
     D wwRChk          S              4A

     D wwRelOk         s              1N   inz(*OFF)    static
     D wwLastRel       s                   like(wwURL)  static
     D wwLastPath      s                   like(pePath) static

     c                   callp     debug_msg('http_long_ParseURL(): ' +
     c                                       'entered')

     c                   eval      peService = *Blanks
     c                   eval      peUserName = *blanks
     c                   eval      pePassword = *blanks
     c                   eval      peHost = *blanks
     c                   eval      pePort = 0
     c                   eval      pePath = *blanks
     c                   eval      wwURL = peURL

      ****************************************************************
      * If this is a relative URL and we have a host & path saved
      * from a previous URL, convert it from a relative to an
      * absolute URL.
      ****************************************************************
     c                   if        wwRelOk = *ON
     c                               and %scan('://': peURL) = 0

     c                   callp     debug_msg('Converting relative URL.')

     c                   if        %subst(peURL:1:1) = '/'
     c                   eval      wwURL = wwLastRel + peURL
     c                   else
     c                   eval      wwURL = wwLastRel + wwLastPath
     c                                   + peURL
     c                   endif

     c                   callp     debug_msg('New URL is ' + wwURL)

     c                   endif

     C****************************************************************
     C*  A valid HTTP url should look like:
     C*      http://www.server.com/somedir/somefile.ext
     C*     https://www.server.com/somedir/somefile.ext
     C*
     C*  and may optionally contain a user name, password & port number:
     C*
     C*     http://user:passwd@www.server.com:80/somedir/somefile.ext
     C****************************************************************

     C* First, extract the URL's "scheme" (which in the case of http
     C*  is the service's name as well):
     c                   eval      wwPos = %scan(':': wwURL)
     c                   if        wwPos < 2 or wwPos > (%len(wwURL)-1)
     c                   callp     SetError(HTTP_BADURL:'Relative URLs '+
     c                              'are not supported!')
     c                   return    -1
     c                   endif

     c                   eval      peService = %subst(wwURL:1:wwPos-1)
     c                   eval      wwURL = %subst(wwURL:wwPos+1)
     c     upper:lower   xlate     peService     peService

      /if defined(HAVE_SSLAPI)
     c                   if        peService<>'http' and peService<>'https'
     c                   callp     SetError(HTTP_BADURL:'Only the http and'+
     c                              ' https protocols are available!')
     c                   return    -1
     c                   endif
      /else
     c                   if        peService<>'http'
     c                   callp     SetError(HTTP_BADURL:'Only the http ' +
     c                              'protocol is available!')
     c                   return    -1
     c                   endif
      /endif

     C* now the URL should be //www.server.com/mydir/somefile.ext!
     C*   make sure it does start with the //, and strip that off.

     c                   if        %subst(wwURL:1:2) <> '//'
     c                   callp     SetError(HTTP_BADURL:'Relative URLs '+
     c                              'are not supported!')
     c                   return    -1
     c                   endif

     c                   eval      wwURL = %subst(wwURL:3)

     C* now, either everything up to the first '/' is part of the
     C*  host name, or the entire string is a hostname.

     c                   eval      wwPos = %scan('/': wwURL)
     c                   if        wwPos = 0
     c                   eval      wwPos = %len(wwURL) + 1
     c                   endif

     c                   eval      peHost = %subst(wwURL:1:wwPos-1)
     c                   if        wwPos > %len(wwURL)
     c                   eval      wwURL = ''
     c                   else
     c                   eval      wwURL = %subst(wwURL:wwPos)
     c                   endif

     C* the host name may optionally contain a user name,
     C*  and possibly also a password:
     c                   eval      wwPos = %scan('@': peHost)
     c                   if        wwPos > 1 and wwPos < %size(peHost)
     c                   eval      wwTemp = %subst(peHost:1:wwPos-1)
     c                   eval      peHost = %subst(peHost:wwPos+1)
     c                   eval      wwPos = %scan(':': wwTemp)
     c                   if        wwPos > 1 and wwPos < %size(wwTemp)
     c                   eval      peUserName = %subst(wwTemp:1:wwPos-1)
     c                   eval      pePassword = %subst(wwTemp:wwPos+1)
     c                   else
     c                   eval      peUserName = wwTemp
     c                   endif
     c                   endif

     C* the host name may also specify a port number:
     c                   eval      wwPos = %scan(':': peHost)
     c                   if        wwPos > 1 and wwPos < %size(peHost)
     c                   eval      wwTemp = %subst(peHost:wwPos+1)
     c                   eval      peHost = %subst(peHost:1:wwPos-1)
     c                   eval      pePort = atoi(%trimr(wwTemp))
     c                   endif

     c* After all that, do we still have a hostname?
     c                   if        peHost=*blanks
     c                   callp     SetError(HTTP_BADURL:'URL does not'+
     c                              ' contain a hostname!')
     c                   return    -1
     c                   endif

     C* Whatever is left should now be the pathname to the file itself.
     C* (or is a parameter or query string for a CGI script)
     c                   eval      pePath = wwURL
     c                   if        %len(pePath)<1 or pePath=*blanks
     c                   eval      pePath = '/'
     c                   endif

      *
      * Save information about this URL so that we can use it to
      * figure out subsequent "relative" URLs.
      *
     c                   eval      wwLastRel = %trimr(peService) + '://'
     c                   if        peUserName <> *blanks
     c                   eval      wwLastRel = wwLastRel
     c                                       + %trim(peUserName)
     c                   if        pePassword <> *blanks
     c                   eval      wwLastRel = wwLastRel + ':'
     c                                       + %trim(pePassword)
     c                   endif
     c                   eval      wwLastRel = wwLastRel + '@'
     c                   endif
     c                   eval      wwLastRel = wwLastRel + %trim(peHost)
     c                   if        pePort <> 0
     c                   eval      wwLastRel = wwLastRel + ':'
     c                                       + %trim(%editc(pePort:'Z'))
     c                   endif

     c                   eval      wwPos = %len(pePath)
     c                   dou       wwPos <= 1
     c                   if        %subst(pePath:wwPos:1) = '/'
     c                   eval      wwLastPath = %subst(pePath:1:wwPos)
     c                   leave
     c                   endif
     c                   eval      wwPos = wwPos -1
     c                   enddo

     c                   eval      wwRelOk = *ON

     c                   return    0
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_url_get(): Retrieve an HTTP document (to a file)
      *
      *       peURL = url to grab  (i.e. http://www.blah.com/dir/file.txt)
      *  peFilename = Filename (in IFS) to save file into
      *  peTimeout  = (optional) give up if no data is received for
      *          this many seconds.
      *  peModTime  = (optiona) only get file if it was changed since
      *          this timestamp.
      *  peContentType = (optional) content type to supply (mainly
      *          useful when talking to CGI scripts)
      *
      *  Returns  -1 upon failure, 0 upon timeout
      *            1 for success, or an HTTP response code
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_url_get    B                   export
     D http_url_get    PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  peFilename                32767A   varying const options(*varsize)
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peModTime                      Z   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwFD            S             10I 0
     D rc              S             10I 0

     c                   callp     debug_msg('http_url_get(): entered')

      *********************************************************
      * open file for writing (O_WRONLY = write only)
      *    if it exists, truncate it (O_TRUNC = truncate)
      *    if it doesnt, create it (O_CREAT = create)
      *    and assign the ASCII codepage to it.
      *********************************************************
     c                   eval      wwFD = open(%trimr(peFilename):
     c                             O_WRONLY+O_TRUNC+O_CREAT+CCSID_OR_CP:
     c                             HTTP_IFSMODE: FILE_CCSID)
     c                   if        wwFD < 0
     c                   callp     SetError(HTTP_FDOPEN:'open(): ' +
     c                               %str(strerror(errno)) )
     c                   return    -1
     c                   endif

      *********************************************************
      *  Call the 'raw' get procedure, telling it to use
      *  the IFS API called 'write' to write data.
      *********************************************************
     c                   select
     c                   when      %parms < 3
     c                   eval      rc = http_url_get_raw(peURL: wwFD:
     c                               %paddr('write'))
     c                   when      %parms < 4
     c                   eval      rc = http_url_get_raw(peURL: wwFD:
     c                               %paddr('write'): peTimeout)
     c                   when      %parms < 5
     c                   eval      rc = http_url_get_raw(peURL: wwFD:
     c                               %paddr('write'): peTimeout:
     c                               peUserAgent)
     c                   when      %parms < 6
     c                   eval      rc = http_url_get_raw(peURL: wwFD:
     c                               %paddr('write'): peTimeout:
     c                               peUserAgent: peModTime)
     c                   when      %parms < 7
     c                   eval      rc = http_url_get_raw(peURL: wwFD:
     c                               %paddr('write'): peTimeout:
     c                               peUserAgent: peModTime: peContentType)
     c                   other
     c                   eval      rc = http_url_get_raw(peURL: wwFD:
     c                               %paddr('write'): peTimeout:
     c                               peUserAgent: peModTime: peContentType:
     c                               peSOAPAction)
     c                   endsl

     c                   callp     close(wwFD)

     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_url_post(): Post data to CGI script and get document
      *
      *         peURL = url to post to (http://www.blah.com/cgi-bin/etc)
      *    pePostData = pointer to data to post to CGI script.
      * pePostDataLen = length of data to post to CGI script.
      *    peFilename = Filename (in IFS) to save file into
      *    peTimeout  = (optional) give up if no data is received for
      *            this many seconds.
      * peContentType = (optional) content type to supply (mainly
      *            useful when talking to CGI scripts)
      *
      *  Returns  -1 upon failure, 0 upon timeout,
      *            1 for success, or an HTTP response code
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_url_post   B                   export
     D http_url_post   PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peFilename                32767A   varying const options(*varsize)
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwFD            S             10I 0
     D rc              S             10I 0

     c                   callp     debug_msg('http_url_post(): entered')

      *********************************************************
      * open file for writing (O_WRONLY = write only)
      *    if it exists, truncate it (O_TRUNC = truncate)
      *    if it doesnt, create it (O_CREAT = create)
      *    and assign the ASCII codepage to it.
      *********************************************************
     c                   eval      wwFD = open(%trimr(peFilename):
     c                             O_WRONLY+O_TRUNC+O_CREAT+CCSID_OR_CP:
     c                             HTTP_IFSMODE: FILE_CCSID)
     c                   if        wwFD < 0
     c                   callp     SetError(HTTP_FDOPEN:'open(): ' +
     c                               %str(strerror(errno)) )
     c                   return    -1
     c                   endif

      *********************************************************
      *  Call the 'raw' post procedure, telling it to use
      *  the IFS API called 'write' to write data.
      *********************************************************
     c                   select
     c                   when      %parms < 5
     c                   eval      rc = http_url_post_raw(peURL:
     c                               pePostData: pePostDataLen:
     c                               wwFD: %paddr('write'))
     c                   when      %parms < 6
     c                   eval      rc = http_url_post_raw(peURL:
     c                               pePostData: pePostDataLen:
     c                               wwFD: %paddr('write'): peTimeout)
     c                   when      %parms < 7
     c                   eval      rc = http_url_post_raw(peURL:
     c                               pePostData: pePostDataLen:
     c                               wwFD: %paddr('write'): peTimeout:
     c                               peUserAgent)
     c                   when      %parms < 8
     c                   eval      rc = http_url_post_raw(peURL:
     c                               pePostData: pePostDataLen:
     c                               wwFD: %paddr('write'): peTimeout:
     c                               peUserAgent: peContentType)
     c                   other
     c                   eval      rc = http_url_post_raw(peURL:
     c                               pePostData: pePostDataLen:
     c                               wwFD: %paddr('write'): peTimeout:
     c                               peUserAgent: peContentType:
     c                               peSOAPAction)
     c                   endsl

     c                   callp     close(wwFD)

     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_url_get_raw(): Retrieve an HTTP document (in raw mode)
      *
      *       peURL = url to grab  (i.e. http://www.blah.com/dir/file.txt)
      *       peFD  = FD to pass back to peProc
      *     peProc  = procedure to call each time data is received.
      *  peTimeout  = (optional) give up if no data is received for
      *          this many seconds.
      *  peModTime  = (optiona) only get file if it was changed since
      *          this timestamp.
      *  peContentType = (optional) content type to supply (mainly
      *          useful when talking to CGI scripts)
      *
      *  Returns  -1 upon failure, 0 upon timeout, or an HTTP response code
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_url_get_raw...
     P                 B                   export
     D http_url_get_raw...
     D                 PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  peFD                         10I 0 value
     D  peProc                         *   value procptr
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peModTime                      Z   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwComm          s               *
     D rc              s             10I 0

      *********************************************************
      * Connect.
      *********************************************************
     c                   if        %parms < 4
     c                   eval      wwComm = http_persist_open( peURL )
     c                   else
     c                   eval      wwComm = http_persist_open( peURL
     c                                                       : peTimeout )
     c                   endif

     c                   if        wwComm = *NULL
     c                   return    -1
     c                   endif

      *********************************************************
      * Get.
      *********************************************************
     c                   select
     c                   when      %parms < 4
     c                   eval      rc = http_persist_get( wwComm
     c                                                  : peURL
     c                                                  : peFD
     c                                                  : peProc )
     c                   when      %parms < 5
     c                   eval      rc = http_persist_get( wwComm
     c                                                  : peURL
     c                                                  : peFD
     c                                                  : peProc
     c                                                  : peTimeout )
     c                   when      %parms < 6
     c                   eval      rc = http_persist_get( wwComm
     c                                                  : peURL
     c                                                  : peFD
     c                                                  : peProc
     c                                                  : peTimeout
     c                                                  : peUserAgent )
     c                   when      %parms < 7
     c                   eval      rc = http_persist_get( wwComm
     c                                                  : peURL
     c                                                  : peFD
     c                                                  : peProc
     c                                                  : peTimeout
     c                                                  : peUserAgent
     c                                                  : peModTime )
     c                   when      %parms < 8
     c                   eval      rc = http_persist_get( wwComm
     c                                                  : peURL
     c                                                  : peFD
     c                                                  : peProc
     c                                                  : peTimeout
     c                                                  : peUserAgent
     c                                                  : peModTime
     c                                                  : peContentType )
     c                   other
     c                   eval      rc = http_persist_get( wwComm
     c                                                  : peURL
     c                                                  : peFD
     c                                                  : peProc
     c                                                  : peTimeout
     c                                                  : peUserAgent
     c                                                  : peModTime
     c                                                  : peContentType
     c                                                  : peSoapAction )
     c                   endsl

      *********************************************************
      * Disconnect.
      *********************************************************
     c                   callp     http_persist_close( wwComm )
     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_url_post_raw(): Post data to CGI script and get document
      *
      *         peURL = url to post to (http://www.blah.com/cgi-bin/etc)
      *    pePostData = pointer to data to post to CGI script.
      * pePostDataLen = length of data to post to CGI script.
      *         peFD  = FD to pass back to peProc
      *       peProc  = procedure to call each time data is received.
      *    peTimeout  = (optional) give up if no data is received for
      *            this many seconds.
      * peContentType = (optional) content type to supply (mainly
      *            useful when talking to CGI scripts)
      *
      *  Returns  -1 upon failure, 0 upon timeout, or an HTTP response code
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_url_post_raw...
     P                 B                   export
     D http_url_post_raw...
     D                 PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peFD                         10I 0 value
     D  peProc                         *   value procptr
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwComm          s               *
     D rc              s             10I 0

      *********************************************************
      * Connect.
      *********************************************************
     c                   if        %parms < 6
     c                   eval      wwComm = http_persist_open( peURL )
     c                   else
     c                   eval      wwComm = http_persist_open( peURL
     c                                                       : peTimeout )
     c                   endif

     c                   if        wwComm = *NULL
     c                   return    -1
     c                   endif

      *********************************************************
      * Post.
      *********************************************************
     c                   select
     c                   when      %parms < 6
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : 0
     c                                                   : *NULL
     c                                                   : pePostData
     c                                                   : pePostDataLen
     c                                                   : peFD
     c                                                   : peProc )

     c                   when      %parms < 7
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : 0
     c                                                   : *NULL
     c                                                   : pePostData
     c                                                   : pePostDataLen
     c                                                   : peFD
     c                                                   : peProc
     c                                                   : peTimeout )

     c                   when      %parms < 8
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : 0
     c                                                   : *NULL
     c                                                   : pePostData
     c                                                   : pePostDataLen
     c                                                   : peFD
     c                                                   : peProc
     c                                                   : peTimeout
     c                                                   : peUserAgent )
     c

     c                   when      %parms < 9
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : 0
     c                                                   : *NULL
     c                                                   : pePostData
     c                                                   : pePostDataLen
     c                                                   : peFD
     c                                                   : peProc
     c                                                   : peTimeout
     c                                                   : peUserAgent
     c                                                   : peContentType )

     c                   when      %parms < 10
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : 0
     c                                                   : *NULL
     c                                                   : pePostData
     c                                                   : pePostDataLen
     c                                                   : peFD
     c                                                   : peProc
     c                                                   : peTimeout
     c                                                   : peUserAgent
     c                                                   : peContentType
     c                                                   : peSOAPAction )
     c                   endsl

      *********************************************************
      * Disconnect.
      *********************************************************
     c                   callp     http_persist_close( wwComm )
     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * http_select_commdriver():  Select & initialize communications
      *    driver.
      *
      *      peCommType = (input) communications type (http/https)
      *
      * Returns pointer to comm driver, or *NULL upon failure
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_select_commdriver...
     P                 B                   export
     D http_select_commdriver...
     D                 PI              *
     D   peCommType                  32A   const

     c                   select
     c                   when      peCommType = 'http'
     c                   eval      p_CommNew = %paddr('COMMTCP_NEW')
      /if defined(HAVE_SSLAPI)
     c                   when      peCommType = 'https'
     c                   eval      p_CommNew = %paddr('COMMSSL_NEW')
      /endif
     c                   other
     c                   callp     SetError( HTTP_NOCDRIV
     c                                     : 'No comm driver to handle '
     c                                     +  %trimr(peCommType)
     c                                     +  ' protocool' )
     c                   return    *NULL
     c                   endsl

     c                   eval      p_CommDriver = comm_new
     c                   return    p_CommDriver
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_url_post_raw2(): Post data to CGI script and get document
      *
      *         peURL = url to post to (http://www.blah.com/cgi-bin/etc)
      *      pePostFD = descriptor number to pass to pePostProc
      *    pePostProc = procedure to call to get POST data.
      *     peDataLen = total length of data that will be sent.
      *      peSaveFD = FD to pass back to peSaveProc
      *    peSaveProc = procedure to call each time data is received.
      *    peTimeout  = (optional) give up if no data is received for
      *            this many seconds.
      * peContentType = (optional) content type to supply (mainly
      *            useful when talking to CGI scripts)
      *
      *  Returns  -1 upon failure, 0 upon timeout, or an HTTP response code
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_url_post_raw2...
     P                 B                   export
     D http_url_post_raw2...
     D                 PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  pePostFD                     10I 0 value
     D  pePostProc                     *   procptr value
     D  peDataLen                    10I 0 value
     D  peSaveFD                     10I 0 value
     D  peSaveProc                     *   value procptr
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwComm          s               *
     D rc              s             10I 0

      *********************************************************
      * Connect.
      *********************************************************
     c                   if        %parms < 7
     c                   eval      wwComm = http_persist_open( peURL )
     c                   else
     c                   eval      wwComm = http_persist_open( peURL
     c                                                       : peTimeout )
     c                   endif

     c                   if        wwComm = *NULL
     c                   return    -1
     c                   endif

      *********************************************************
      * Post.
      *********************************************************
     c                   select
     c                   when      %parms < 7
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : pePostFD
     c                                                   : pePostProc
     c                                                   : *NULL
     c                                                   : peDataLen
     c                                                   : peSaveFD
     c                                                   : peSaveProc )

     c                   when      %parms < 8
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : pePostFD
     c                                                   : pePostProc
     c                                                   : *NULL
     c                                                   : peDataLen
     c                                                   : peSaveFD
     c                                                   : peSaveProc
     c                                                   : peTimeout  )

     c                   when      %parms < 9
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : pePostFD
     c                                                   : pePostProc
     c                                                   : *NULL
     c                                                   : peDataLen
     c                                                   : peSaveFD
     c                                                   : peSaveProc
     c                                                   : peTimeout
     c                                                   : peUserAgent )

     c                   when      %parms < 10
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : pePostFD
     c                                                   : pePostProc
     c                                                   : *NULL
     c                                                   : peDataLen
     c                                                   : peSaveFD
     c                                                   : peSaveProc
     c                                                   : peTimeout
     c                                                   : peUserAgent
     c                                                   : peContentType )

     c                   other
     c                   eval      rc = http_persist_post( wwComm
     c                                                   : peURL
     c                                                   : pePostFD
     c                                                   : pePostProc
     c                                                   : *NULL
     c                                                   : peDataLen
     c                                                   : peSaveFD
     c                                                   : peSaveProc
     c                                                   : peTimeout
     c                                                   : peUserAgent
     c                                                   : peContentType
     c                                                   : peSOAPAction )
     c                   endsl

      *********************************************************
      * Disconnect.
      *********************************************************
     c                   callp     http_persist_close( wwComm )
     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_url_post_stmf(): Post data to CGI script from stream file
      *
      *         peURL = url to post to (http://www.blah.com/cgi-bin/etc)
      *    pePostFile = Filename (in IFS) of file to send to http server
      *    peRecvFile = Filename (in IFS) of stream file containing reply
      *    peTimeout  = (optional) give up if no data is received for
      *            this many seconds.
      * peContentType = (optional) content type to supply (mainly
      *            useful when talking to CGI scripts)
      *
      *  Returns  -1 upon failure, 0 upon timeout,
      *            1 for success, or an HTTP response code
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_url_post_stmf...
     P                 B                   export
     D http_url_post_stmf...
     D                 PI            10I 0
     D  peURL                     32767A   varying const options(*varsize)
     D  pePostFile                32767A   varying const options(*varsize)
     D  peRecvFile                32767A   varying const options(*varsize)
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwPostFD        S             10I 0
     D wwRecvFD        S             10I 0
     D wwDataSize      s             10U 0
     D rc              S             10I 0
     D wwStat          s                   like(statds)

     c                   callp     debug_msg('http_url_post_stmf(): ' +
     c                                      'entered')

      *********************************************************
      * open file to be posted
      *********************************************************
     c                   callp     debug_msg('getting post file size...')
     c                   if        stat(pePostFile: %addr(wwStat)) < 0
     c                   callp     SetError(HTTP_FDSTAT:'stat(): ' +
     c                                        %str(strerror(errno)) )
     c                   return    -1
     c                   endif

     c                   eval      p_statds = %addr(wwStat)
     c                   eval      wwDataSize = st_size

     c                   callp     debug_msg('opening file to be sent...')
     c                   eval      wwPostFD = open(pePostFile : O_RDONLY)
     c                   if        wwPostFD < 0
     c                   callp     SetError(HTTP_FDOPEN:'open(): ' +
     c                                        %str(strerror(errno)) )
     c                   return    -1
     c                   endif

      *********************************************************
      * open file for writing (O_WRONLY = write only)
      *    if it exists, truncate it (O_TRUNC = truncate)
      *    if it doesnt, create it (O_CREAT = create)
      *    and assign the ASCII codepage to it.
      *********************************************************
     c                   callp     debug_msg('opening file to be received')
     c                   eval      wwRecvFD = open( peRecvFile
     c                                            : O_WRONLY
     c                                            + O_TRUNC
     c                                            + O_CREAT
     c                                            + CCSID_OR_CP
     c                                            : HTTP_IFSMODE
     c                                            : FILE_CCSID
     c                                            )
     c                   if        wwRecvFD < 0
     c                   callp     SetError(HTTP_FDOPEN:'open(): ' +
     c                               %str(strerror(errno)) )
     c                   callp     close(wwPostFD)
     c                   return    -1
     c                   endif

      *********************************************************
      *  Call the 'raw' post procedure, telling it to use
      *  the IFS API called 'write' to write data.
      *********************************************************
     c                   select
     c                   when      %parms < 4
     c                   eval      rc = http_url_post_raw2( peURL
     c                                                    : wwPostFD
     c                                                    : %paddr('read')
     c                                                    : wwDataSize
     c                                                    : wwRecvFD
     c                                                    : %paddr('write'))
     c                   when      %parms < 5
     c                   eval      rc = http_url_post_raw2( peURL
     c                                                    : wwPostFD
     c                                                    : %paddr('read')
     c                                                    : wwDataSize
     c                                                    : wwRecvFD
     c                                                    : %paddr('write')
     c                                                    : peTimeout )
     c                   when      %parms < 6
     c                   eval      rc = http_url_post_raw2( peURL
     c                                                    : wwPostFD
     c                                                    : %paddr('read')
     c                                                    : wwDataSize
     c                                                    : wwRecvFD
     c                                                    : %paddr('write')
     c                                                    : peTimeout
     c                                                    : peUserAgent )
     c                   when      %parms < 7
     c                   eval      rc = http_url_post_raw2( peURL
     c                                                    : wwPostFD
     c                                                    : %paddr('read')
     c                                                    : wwDataSize
     c                                                    : wwRecvFD
     c                                                    : %paddr('write')
     c                                                    : peTimeout
     c                                                    : peUserAgent
     c                                                    : peContentType )
     c                   other
     c                   eval      rc = http_url_post_raw2( peURL
     c                                                    : wwPostFD
     c                                                    : %paddr('read')
     c                                                    : wwDataSize
     c                                                    : wwRecvFD
     c                                                    : %paddr('write')
     c                                                    : peTimeout
     c                                                    : peUserAgent
     c                                                    : peContentType
     c                                                    : peSOAPAction )
     c                   endsl

     c                   callp     close(wwPostFD)
     c                   callp     close(wwRecvFD)
     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_persist_open(): Open a persistent HTTP session
      *
      *       peURL = url to connect to
      *  peTimeout  = (optional) give up if no data is received for
      *          this many seconds.
      *
      *  Returns *NULL upon failure, or
      *          pointer to HTTP communication session
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_persist_open...
     P                 B                   export
     D http_persist_open...
     D                 PI              *
     D  peURL                     32767A   varying const options(*varsize)
     D  peTimeout                    10I 0 value options(*nopass)

     D wwTimeout       S             10I 0
     D wwServ          S             32A
     D wwUser          S             32A
     D wwPass          S             32A
     D wwHost          S            256A
     D wwPort          S             10I 0
     D wwOrigHost      S            256A
     D wwOrigPort      S             10I 0
     D wwPath          S          32767A   varying
     D wwComm          s               *
     D p_addr          s               *

     c                   callp     debug_msg('http_persist_open():'
     c                                     + ' entered')

      *********************************************************
      * Set up optional parameters (timeout)
      *********************************************************
     c                   if        %parms>=2 and peTimeout>=0
     c                   eval      wwTimeout = peTimeout
     c                   else
     c                   eval      wwTimeout = HTTP_TIMEOUT
     c                   endif

      *********************************************************
      *  Parse URL into it's components
      *********************************************************
     c                   if        http_long_ParseURL(peURL: wwServ: wwUser:
     c                                wwPass: wwHost: wwPort: wwPath) < 0
     c                   return    *NULL
     c                   endif

      *********************************************************
      *  Proxy address provided?
      *********************************************************
     c                   eval      wwOrigHost = wwHost
     c                   eval      wwOrigPort = wwPort

     c                   if        dsProxyHost <> *blanks
     c                   eval      wwHost = dsProxyHost
     c                   endif

     c                   if        dsProxyPort <> *zeros
     c                   eval      wwPort = dsProxyPort
     c                   endif

     c                   if        dsProxyHost <> *blanks
     c                               and wwServ='https'
     c                   eval      dsProxyTun = *ON
     c                   else
     c                   eval      dsProxyTun = *OFF
     c                   endif

      *********************************************************
      *  Select comm driver & build a socket address structure
      *********************************************************
     c                   eval      wwComm = http_select_commdriver(wwServ)
     c                   if        wwComm = *NULL
     c                   return    *NULL
     c                   endif

     c                   eval      p_addr = comm_resolve( wwComm
     c                                                  : %trimr(wwHost)
     c                                                  : %trimr(wwServ)
     c                                                  : wwPort
     c                                                  : *OFF    )

     c                   if        p_addr = *NULL
     c                   return    *NULL
     c                   endif

      *********************************************************
      * Connect to server
      *********************************************************
     c                   if        comm_Connect( wwComm
     c                                         : p_addr
     c                                         : wwTimeout
     c                                         ) = *OFF
     c                   return    *NULL
     c                   endif

      *********************************************************
      * Establish proxy tunnelling if necessary
      *********************************************************
     c                   if        dsProxyTun = *ON
     c                   if        proxy_tunnel( wwComm
     c                                         : wwServ
     C                                         : wwOrigHost
     c                                         : wwOrigPort
     c                                         : wwTimeout ) <> 0
     c                   callp     http_persist_close( wwComm )
     c                   return    *NULL
     c                   endif
     c                   endif

      *********************************************************
      * Upgrade connection security (if driver supports it)
      * (if wwComm is the SSL driver, this starts SSL encryption)
      *********************************************************
     c                   if        comm_Upgrade( wwComm
     c                                         : wwTimeout
     c                                         ) = *OFF
     c                   return    *NULL
     c                   endif

     c                   return    wwComm
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_persist_get(): Get a file using a persistent HTTP session
      *
      *         peComm = (input) pointer to persistent HTTP comm session
      *          peURL = (input) URL to get from persistent HTTP
      *           peFD = (input) FD to pass back to peProc
      *         peProc = (input) procedure to call each time data is
      *                          received.
      *      peTimeout = (input/optional) time-out when no data is received
      *                          for this many seconds.
      *      peModTime = (input/optional) only get file if it was changed
      *                          since this timestamp.
      *  peContentType = (input/optional) content type to supply (mainly
      *                          useful when talking to CGI scripts)
      *   peSOAPAction = (input/optional) string to send in the SOAPAction:
      *                          HTTP header when making a SOAP request.
      *
      *  Returns  1 if successful,
      *           0 if timed out
      *          -1 if an internal error occurs
      *          or an HTTP response code if an error comes from the server
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_persist_get...
     P                 B                   export
     D http_persist_get...
     D                 PI            10I 0
     D  peComm                         *   value
     D  peURL                     32767A   varying const options(*varsize)
     D  peFD                         10I 0 value
     D  peProc                         *   value procptr
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peModTime                      Z   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwUA            S             64A
     D wwMT            S               Z
     D wwCT            S             64A
     D wwSA            S             64A
     D wwTimeout       S             10I 0
     D p_UserAgent     S               *
     D wwUserAgent     S             64A   based(p_UserAgent)
     D p_ModTime       S               *
     D wwModTime       S               Z   based(p_ModTime)
     D p_ContentType   S               *
     D wwContentType   S             64A   based(p_ContentType)
     D p_SOAPAction    S               *
     D wwSOAPAction    S             64A   based(p_SOAPAction)
     D wwServ          S             32A
     D wwUser          S             32A
     D wwPass          S             32A
     D wwHost          S            256A
     D wwPort          S             10I 0
     D wwPath          S          32767A   varying
     D rc              S             10I 0
     D wwSecure        s              1N   inz(*OFF)

     c                   callp     debug_msg('http_persist_get():'
     c                                     + ' entered')

      *********************************************************
      * Set up optional parameters
      *********************************************************
      * If no timeout given, default to 5 minutes
     c                   if        %parms>=5 and peTimeout>=0
     c                   eval      wwTimeout = peTimeout
     c                   else
     c                   eval      wwTimeout = HTTP_TIMEOUT
     c                   endif

      * If no user-agent given, pass '*OMIT' to next proc
     c                   if        %parms >= 6
     c                   eval      p_UserAgent = %addr(wwUA)
     c                   eval      wwUserAgent = peUserAgent
     c                   else
     c                   eval      p_UserAgent = *NULL
     c                   endif

      * If no mod time given, pass '*OMIT' to next proc
     c                   if        %parms >= 7
     c                   eval      p_ModTime = %addr(wwMT)
     c                   eval      wwModTime = peModTime
     c                   else
     c                   eval      p_ModTime = *NULL
     c                   endif

      * If no content-type given, pass '*OMIT' to next proc
     c                   if        %parms >= 8
     c                   eval      p_ContentType = %addr(wwCT)
     c                   eval      wwContentType = peContentType
     c                   else
     c                   eval      p_ContentType = *NULL
     c                   endif

      * If no SoapAction given, pass '*OMIT' to next proc
     c                   if        %parms >= 9
     c                   eval      p_SOAPAction = %addr(wwSA)
     c                   eval      wwSOAPAction = peSOAPAction
     c                   else
     c                   eval      p_SOAPAction = *NULL
     c                   endif

      *********************************************************
      * Parse URL.  We have to do this again, despite having
      *   done it in HTTP_persist_open() because it's
      *   possible to change some parts of it from call to
      *   call.
      *********************************************************
     c                   if        http_long_ParseURL(peURL: wwServ: wwUser:
     c                                wwPass: wwHost: wwPort: wwPath) < 0
     c                   return    -1
     c                   endif

     c                   if        wwServ='https'
     c                   eval      wwSecure = *ON
     c                   endif

     c                   callp     setUrlAuth(wwUser: wwPass)

      *********************************************************
      *  Perform the GET operation
      *********************************************************
     c                   eval      rc = do_get( peProc
     c                                        : peComm
     c                                        : peFD
     c                                        : wwTimeout
     c                                        : wwPath
     c                                        : wwHost
     c                                        : wwModTime
     c                                        : wwUserAgent
     c                                        : wwContentType
     c                                        : wwSoapAction
     c                                        : wwPort
     c                                        : wwSecure
     c                                        : wwServ
     c                                        )

     c                   if        wkSaveAuth <> *blanks
     c                   eval      dsAuth = wkSaveAuth
     c                   eval      wkSaveAuth = *blanks
     c                   endif

     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_persist_post(): Post data to CGI script and get document
      *                       using a persistent connection
      *
      *         peComm = (input) pointer to persistent HTTP comm session
      *          peURL = (input) URL to post to with persistent HTTP
      * --------
      *       pePostFD = (input) Opaque integer to pass to pePostProc
      *     pePostProc = (input) Pointer to call-back procedure for
      *                          posting data to server.  If you pass
      *                          *NULL for this, you should use pePostData
      *                          instead.
      * -- or --
      *     pePostData = (input) Pointer to data to post.  If you pass
      *                          *NULL for this, you should use pePostProc
      *                          instead.
      * --------
      *  pePostDataLen = (input) Total length, in bytes, of post data.
      *       peSaveFD = (input) Opaque integer passed to peSaveProc
      *     peSaveProc = (input) Pointer to call-back procedure that is
      *                          called when data is received from HTTP
      *                          server.
      *      peTimeout = (input/optional) time-out when no data is received
      *                          for this many seconds.
      *  peContentType = (input/optional) content type to supply (mainly
      *                          useful when talking to CGI scripts)
      *   peSOAPAction = (input/optional) string to send in the SOAPAction:
      *                          HTTP header when making a SOAP request.
      *
      *  Returns  1 if successful,
      *           0 if timed out
      *          -1 if an internal error occurs
      *          or an HTTP response code if an error comes from the server
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_persist_post...
     P                 B                   export
     D http_persist_post...
     D                 PI            10I 0
     D  peComm                         *   value
     D  peURL                     32767A   varying const options(*varsize)
     D  pePostFD                     10I 0 value
     D  pePostProc                     *   value procptr
     D  pePostData                     *   value
     D  pePostDataLen                10I 0 value
     D  peSaveFD                     10I 0 value
     D  peSaveProc                     *   value procptr
     D  peTimeout                    10I 0 value options(*nopass)
     D  peUserAgent                  64A   const options(*nopass:*omit)
     D  peContentType                64A   const options(*nopass:*omit)
     D  peSOAPAction                 64A   const options(*nopass:*omit)

     D wwUA            S             64A
     D wwCT            S             64A
     D wwSA            S             64A
     D wwTimeout       S             10I 0
     D p_UserAgent     S               *
     D wwUserAgent     S             64A   based(p_UserAgent)
     D p_ContentType   S               *
     D wwContentType   S             64A   based(p_ContentType)
     D p_SOAPAction    S               *
     D wwSOAPAction    S             64A   based(p_SOAPAction)
     D wwServ          S             32A
     D wwUser          S             32A
     D wwPass          S             32A
     D wwHost          S            256A
     D wwPort          S             10I 0
     D wwPath          S          32767A   varying
     D rc              S             10I 0
     D wwSendProc      s               *   procptr
     D wwSecure        s              1N   inz(*OFF)
     D wwPostData      s                   like(pePostData) inz(*null)
     D wwPostLen       s                   like(pePostDataLen)

     c                   callp     debug_msg('http_persist_post(): entered')

      *********************************************************
      *  Handle the optional parameters
      *********************************************************
      * If no timeout, default to 5 mins
     c                   if        %parms>=9 and peTimeout>=0
     c                   eval      wwTimeout = peTimeout
     c                   else
     c                   eval      wwTimeout = HTTP_TIMEOUT
     c                   endif

      * If no user-agent, pass '*OMIT' to next proc
     c                   if        %parms>=10
     c                   eval      p_UserAgent = %addr(wwUA)
     c                   eval      wwUserAgent = peUserAgent
     c                   else
     c                   eval      p_UserAgent = *NULL
     c                   endif

      * If no content-type, pass '*OMIT' to next proc
     c                   if        %parms>=11
     c                   eval      p_ContentType = %addr(wwCT)
     c                   eval      wwContentType = peContentType
     c                   else
     c                   eval      p_ContentType = *NULL
     c                   endif

      * If no SOAPaction, pass '*OMIT' to next proc
     c                   if        %parms>=12
     c                   eval      p_SOAPAction = %addr(wwSA)
     c                   eval      wwSOAPAction = peSOAPAction
     c                   else
     c                   eval      p_SOAPAction = *NULL
     c                   endif

      *********************************************************
      *  Parse URL into it's components, and then look up
      *  host & build a socket address structure:
      *********************************************************
     c                   if        http_long_ParseURL(peURL: wwServ: wwUser:
     c                                wwPass: wwHost: wwPort: wwPath) < 0
     c                   return    -1
     c                   endif

     c                   if        wwServ = 'https'
     c                   eval      wwSecure = *ON
     c                   endif

     c                   callp     setUrlAuth(wwUser: wwPass)

      *********************************************************
      * Translate POST data to ASCII (if necessary)
      *********************************************************
     c                   eval      wwPostLen = pePostDataLen
     c                   if        pePostDataLen>0 and pePostProc=*NULL
     c                   eval      wwPostLen = http_xlatedyn( pePostDataLen
     c                                                      : pePostData
     c                                                      : TO_ASCII
     c                                                      : wwPostData )
     c                   if        wwPostLen = -1
     c                   return    -1
     c                   endif
     c                   endif

      *********************************************************
      * Perform the POST operation
      *********************************************************
     c                   if        pePostProc = *NULL
     c                   eval      wwSendProc = %paddr('SENDDOC')
     c                   else
     c                   eval      wwSendProc = %paddr('SENDRAW')
     c                   eval      wwPostData = *NULL
     c                   endif

     c                   eval      rc = do_post( peSaveProc
     c                                         : wwSendProc
     c                                         : wwPostData
     c                                         : wwPostLen
     c                                         : peComm
     c                                         : peSaveFD
     c                                         : wwTimeout
     c                                         : wwPath
     c                                         : wwHost
     c                                         : wwUserAgent
     c                                         : wwContentType
     c                                         : wwSOAPAction
     c                                         : pePostProc
     c                                         : pePostFD
     c                                         : wwPort
     c                                         : wwSecure
     c                                         : wwServ
     c                                         )

     c                   if        wkSaveAuth <> *blanks
     c                   eval      dsAuth = wkSaveAuth
     c                   eval      wkSaveAuth = *blanks
     c                   endif

     c                   if        wwPostData <> *Null
     c                   callp     xdealloc(wwPostData)
     c                   endif

     c                   return    rc
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_persist_close(): End a persistent HTTP session
      *
      *     peComm = (input) pointer to persistent HTTP comm session
      *
      *  returns 0 if successful, -1 otherwise
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_persist_close...
     P                 B                   export
     D http_persist_close...
     D                 PI            10I 0
     D  peComm                         *   value
     c                   return    http_close( 0 : peComm)
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * parse important fields from the response chain
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P parse_resp_chain...
     P                 B
     D parse_resp_chain...
     D                 PI            10I 0
     D  peRespChain               32767A   varying const
     D  peRC                         10I 0
     D  peTE                         32A
     D  peCLen                       10u 0
     D  peUseCL                       1N
     D  peAuthErr                     1N
     D  peProxyAuthErr...
     D                                1N
     D  peHost                      256A   varying const
     D  pePath                      256A   varying const

     D wwChain         S          32767A   varying
     D wwPos           S             10I 0
     D wwPos2          S             10I 0
     D wwCL            S             32A
     D LF              S              1A   inz(x'25') static
     D CRLF            S              2A   inz(x'0d25') static

     c                   select
     c                   when      peRC = 401
     c                   eval      peAuthErr = *on

     c                   when      peRC = 407
     c                   eval      peProxyAuthErr = *on

     c                   when      peRC<=100 or peRC=204 or peRC=304
     c                   return    peRC
     c                   endsl

      *********************************************************
      * Make an uppercase copy of the response chain so
      * we can do case-insensitive searching on it.
      *********************************************************
     c                   eval      wwChain = peRespChain
     c     upper:lower   xlate     wwChain       wwChain

      *********************************************************
      * parse out transfer-encoding:
      *********************************************************
     c                   eval      peTE = 'identity'
     c                   eval      wwPos = %scan(LF+'transfer-encoding:':
     c                                  wwChain)
     c                   if        wwPos > 0
     c                   eval      wwPos = wwPos + 19
     c                   eval      wwPos2 = %scan(CRLF:wwChain:wwPos)
     c                   eval      wwPos2 = wwPos2 - wwPos
     c                   if        wwPos2 > 0
     c                   eval      peTE= %trim(%subst(wwChain:wwPos:wwPos2))
     c                   endif
     c                   endif

     c                   if        %scan('chunked': peTE)=0  and
     c                               %scan('identity': peTE)=0
     c                   callp     SetError(HTTP_XFRENC: 'The "' +
     c                              %trim(peTE) + '" transfer encoding is'+
     c                              ' not supported.')
     c                   return    -1
     c                   endif

      *********************************************************
      *  parse out content-length if using "identity,"
      *   (this is irrelevant for "chunked")
      *********************************************************
     c                   eval      wwCL = '0'
     c                   eval      peUseCL = *OFF
     c                   if        %scan('identity': peTE) > 0
     c                   eval      wwPos = %scan(LF+'content-length:':
     c                                           wwChain)
     c                   if        wwPos > 0
     c                   eval      wwPos = wwPos + 16
     c                   eval      wwPos2 = %scan(CRLF:wwChain:wwPos)
     c                   eval      wwPos2 = wwPos2 - wwPos
     c                   if        wwPos2 > 0
     c                   eval      wwCL= %trim(%subst(wwChain:wwPos:wwPos2))
     c                   eval      peUseCL = *ON
     c                   endif
     c                   endif
     c                   endif
     c                   eval      peCLen = atoll(wwCL)

     c                   callp     debug_msg('recvdoc parms: '+%trim(peTE)+
     c                                ' ' + %trim(%editc(peCLen:'P')))

      *********************************************************
      * parse out www-authenticate: header:
      * (used when userid/password for actual site is required)
      *********************************************************
     c                   eval      wwPos=%scan('www-authenticate:':wwChain)
     c                   if        wwPos > 0
     c                   callp     interpret_auth(peRespChain: wwPos)
     c                   endif

     c                   if        peAuthErr = *On
     c                   callp     SetError(HTTP_NDAUTH:'This page requires' +
     c                               ' a user-id & password')
     c                   return    -1
     c                   endif

      *********************************************************
      * parse proxy-authenticate: header
      * (used when userid/password for proxy is required)
      * this header may occur more than once (especially on
      * MS ISA servers, where Kerberos, NTLM, Basic, etc.
      * may be provided)
      *********************************************************
     c                   eval      dsProxyAuthBasic = *off
     c                   eval      dsProxyAuthRealm = *blanks
     c                   eval      wwPos=%scan(LF+'proxy-authenticate:':wwChain)
     c                   dow       wwPos > 0
     c                   callp     interpret_proxy_auth(peRespChain: wwPos)
     c                   eval      wwPos=%scan(LF+'proxy-authenticate:':wwChain:
     c                               wwPos+1)
     c                   enddo

     c                   if        peProxyAuthErr = *On
     c                   callp     SetError(HTTP_PXNDAUTH:'This proxy ' +
     c                               ' requires a user-id & password')
     c                   return    -1
     c                   endif

      *********************************************************
      * parse out 'location:' header:
      *  (used for redirects)
      *********************************************************
     c                   eval      %len(wkRedirLoc) = 0
     c                   eval      wwPos = %scan(LF+'location:':
     c                                           wwChain)
     c                   if        wwPos > 0
     c                   eval      wwPos = wwPos + 11
     c                   eval      wwPos2 = %scan(CRLF:wwChain:wwPos)
     c                   eval      wwPos2 = wwPos2 - wwPos
     c                   if        wwPos2 > 0
     c                   eval      wkRedirLoc =
     c                             %trim(%subst(peRespChain:wwPos:wwPos2))
     c                   endif
     c                   endif

      *********************************************************
      * Call HTTP header parsing module
      *********************************************************
     c                   callp     header_parse(peRespChain: *NULL)

     c                   if        use_cookies = *on
     c                   callp     header_load_cookies(peHost: pePath)
     c                   endif

      *********************************************************
      * Call user supplied parsing proc if available
      *********************************************************
     c                   if        wkParseHdrProc <> *NULL
     c                   callp     parse_hdrs(peRespChain: wkParseHdrdata)
     c                   endif
     c                   if        wkParseHdrLong <> *NULL
     c                   callp     parse_hdr_long( peRespChain
     c                                           : wkParseHdrLongData )
     c                   endif

     c                   return    peRC
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * http_set_100_timeout(): Set value for 100-continue timeouts.
      *
      * This API lets you change the timeout value that HTTPAPI waits
      * for a "100 Continue" response from the server during a POST
      * request. Some servers send this response, others do not. By
      * default, HTTPAPI waits 0.5 seconds for a response, and if it's
      * not received, sends the POST.
      *
      *   Set this to a low value (0.1?) when working with a server
      *   that does not send a "100 continue" response. That way, you
      *   don't have to wait for the timeout.
      *
      *   Set it to a higher value (10.0?) when working with a server
      *   whose responses are slower, but will definitely send the
      *   "100".  That way, HTTPAPI won't give up on the response and
      *   get out of sync.
      *
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_set_100_timeout...
     P                 B                   export
     D http_set_100_timeout...
     D                 PI
     D peTimeout                     10P 3 value
     c                   eval      wk100time = peTimeout
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * http_use_cookies(): Turns on/off HTTPAPI's cookie parsing and
      *                     caching routines.
      *
      *    peSetting = (input) *ON = HTTPAPI will read and send cookies
      *                       *OFF = HTTPAPI will ignore cookies
      *                              (has no affect on cookies supplied
      *                               via an exit procedure)
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_use_cookies...
     P                 B                   export
     D http_use_cookies...
     D                 PI
     D   peSetting                    1N   const
     c                   eval      use_cookies = peSetting
     P                 E

      ****************************************************************
      * Proxy support                                                *
      ****************************************************************

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_proxy_setauth():   Set HTTP Proxy Authentication Information
      *
      *     peAuthType = Authentication Type (HTTP_AUTH_NONE or
      *                     HTTP_AUTH_BASIC)
      *     peUsername = UserName to use
      *     pePasswd   = Password to use
      *
      *  Returns -1 upon error, or 0 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_proxy_setauth...
     P                 B                   export
     D http_proxy_setauth...
     D                 PI            10I 0
     D   peAuthType                   1A   const
     D   peUsername                  80A   const
     D   pePasswd                  1024A   const

     D wwString        S           1105A
     D wwEncoded       S           1476A
     D wwEncLen        S             10I 0

     c                   callp     debug_msg('http_proxy_setauth(): entered')
     c                   if        peAuthType<>HTTP_AUTH_BASIC
     c                               and peAuthType<>HTTP_AUTH_NONE
     c                   callp     SetError(HTTP_ATHTYP: 'Invalid authenti'+
     c                                   'cation type!')
     c                   return    -1
     c                   endif

     C*************************************************
     c* Calculate strings for NO authentication
     C*************************************************
     c                   if        peAuthType = HTTP_AUTH_NONE
     c                   eval      dsProxyAuthType = HTTP_AUTH_NONE
     c                   eval      dsProxyAuthStr = ''
     c                   return    0
     c                   endif

     c                   if        %scan(':':peUserName) > 0
     c                              or %scan(':': pePasswd) > 0
     c                   callp     SetError(HTTP_ATHVAL: 'HTTP Proxy Auth '+
     c                               'values cannot contain a colon!')
     c                   return    -1
     c                   endif

     c                   eval      dsProxyAuthUser = peUserName
     c                   eval      dsProxyAuthPasswd = pePasswd

     c                   eval      wwString = %trimr(peUserName) + ':' +
     c                                        %trimr(pePasswd)
     c                   callp     http_xlate( %len(%trimr(wwString))
     c                                       : wwString
     c                                       : TO_ASCII )

     c                   eval      wwEncLen = base64_encode(%addr(wwString):
     c                                             %len(%trimr(wwString)):
     c                                             %addr(wwEncoded):
     c                                             %size(wwEncoded))

     c                   eval      dsProxyAuthStr = %subst(wwEncoded:1:wwEncLen)
     c                   eval      dsProxyAuthType = peAuthType

     c                   return    0
     P                 E

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_setproxy():   Set HTTP Proxy Address
      *
      *     peHost = Proxy host name
      *     psPort = Proxy port
      *
      *  Returns -1 upon error, or 0 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_setproxy   B                   export
     D http_setproxy   PI            10I 0
     D   peHost                     256A   const
     D   pePort                      10I 0 const

     c                   eval      dsProxyHost = peHost
     c                   eval      dsProxyPort = pePort
     c                   eval      dsProxyTun  = *off

     c                   return    0

     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      *  http_proxy_getauth():   Get HTTP Proxy Authentication Information
      *
      *   Call this proc after you receive a HTTP_PXNDAUTH error
      *   to determine the authentication credentials that are required
      *
      *  The following parms are returned to your program:
      *
      *     peBasic = *ON if BASIC auth is allowed
      *     peRealm = Auth realm.  Present this to the user to identify
      *               which password you're looking for.  For example
      *               if peRealm is "secureproxy.com" you might say
      *               "enter password for secureproxy.com" to user.
      *
      *   After getting the userid & password from the user (or database)
      *   you'll need to call http_proxy_setauth()
      *
      *  Returns -1 upon error, or 0 if successful
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P http_proxy_getauth...
     P                 B                   export
     D http_proxy_getauth...
     D                 PI            10I 0
     D   peBasic                      1N
     D   peRealm                    124A

     c                   callp     debug_msg('http_proxy_getauth(): entered')
     c                   if        dsProxyAuthRealm = *blanks
     c                   callp     SetError(HTTP_NOAUTH: 'Proxy did ' +
     c                              'not ask for authentication!')
     c                   return    -1
     c                   endif

     c                   eval      peBasic = dsProxyAuthBasic
     c                   eval      peRealm = dsProxyAuthRealm

     c                   return    0
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * proxy_tunnel(): This establishes a tunnel through a proxy,
      *                 this lets us communicate directly with the
      *                 destination HTTP server (required for SSL)
      *
      *    peComm = Communication driver
      *    peServ = Service (http or https)
      *    peHost = destination (not proxy) host
      *    pePort = destination port number
      * peTimeout = timeout
      *
      * returns 0 if successful, otherwise an HTTP response number
      *         or -1 upon internal error
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P proxy_tunnel    B
     D proxy_tunnel    PI            10I 0
     D   peComm                        *   value
     D   peServ                      32a   const
     D   peHost                     256a   const
     D   pePort                      10i 0 value
     D   peTimeout                   10i 0 value

     D wwReq           S          32767A   varying
     D wwResp          s          32767A   varying

     D rc              S             10I 0
     D wwPort          S             10I 0
     D wwTE            S             32A
     D wwClen          s             10U 0
     D wwUseCL         s              1N
     D wwAuthErr       S              1N   inz(*OFF)
     D wwProxyAuthErr...
     D                 S              1N   inz(*OFF)

      *************************************************
      * Create proxy tunnelling request string
      *************************************************
     c                   if        pePort=0
     c                   if        peServ='https'
     c                   eval      wwPort=443
     c                   else
     c                   eval      wwPort=80
     c                   endif
     c                   endif

     c                   eval      wwReq = 'CONNECT ' + %trim(peHost)
     c                                   + ':' + %trim(%editc(wwPort:'L'))
     c                                   + ' HTTP/1.1'
     C                                   + CRLF

     c                   if        pePort = 0
     c                   eval      wwReq = wwReq
     C                                   + 'Host: ' + %trim(peHost)
     c                                   + ' HTTP/1.1'
     C                                   + CRLF
     c                   else
     c                   eval      wwReq = wwReq
     C                                   + 'Host: ' + %trim(peHost)
     c                                   + ':' + %trim(%editc(pePort:'L'))
     c                                   + ' HTTP/1.1'
     C                                   + CRLF
     c                   endif

     c                   eval      wwReq = wwReq
     C                                   + 'User-Agent: ' + HTTP_USERAGENT
     C                                   + CRLF
     c                                   + 'Proxy-Connection: keep-alive'
     c                                   + CRLF

     c                   if        dsProxyAuthType = HTTP_AUTH_BASIC
     c                   eval      wwReq = wwReq
     c                                   + 'Proxy-Authorization: Basic '
     c                                   + dsProxyAuthStr + CRLF
     c                   endif

     C                   eval      wwReq = wwReq + CRLF

      *************************************************
      * Send request to proxy server
      *************************************************
     c                   eval      rc = SendReq( peComm
     c                                         : %addr(wwReq)+2
     c                                         : %len(wwReq)
     c                                         : peTimeout )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

      *************************************************
      * Receive & parse proxy server's response.
      *************************************************
     c                   eval      rc = RecvResp( peComm
     c                                          : wwResp
     c                                          : %size(wwResp)
     c                                          : peTimeout
     c                                          : *OFF )
     c                   if        rc < 1
     c                   return    rc
     c                   endif

     c                   eval      rc = parse_resp_chain( wwResp
     c                                                  : rc
     c                                                  : wwTE
     c                                                  : wwCLen
     c                                                  : wwUseCL
     c                                                  : wwAuthErr
     c                                                  : wwProxyAuthErr
     c                                                  : %trim(dsProxyHost)
     c                                                  : '/' )
     c                   if        rc <> 200
     c                   return    rc
     c                   endif

     c                   return    0
     P                 E


      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * wkDebugNoClose(): Used to disable call to debug_close.
      *
      * This procedures are for internal use only. Used by the following
      * procedures to disable closing the debug file in http_close():
      *
      *   http_url_get_xml       -> http_url_get_raw   -> persist_close -> http_close
      *   http_url_post_xml      -> http_url_post_raw  -> persist_close -> http_close
      *   http_url_post_stmf_xml -> http_url_post_raw2 -> persist_close -> http_close
      *   http_get_xmltf         -> http_url_get       -> http_url_get_raw   -> ...
      *   http_post_xmltf        -> http_url_post      -> http_url_post_raw  -> ...
      *   http_post_stmf_xmltf   -> http_url_post_stmf -> http_url_post_raw2 -> ...
      *
      * The procedures above perform additional processing after having
      * done the http stuff and hence need the debug file. If the debug file
      * is closed by http_close() and something goes wrong after that, the
      * debug file is reopened and truncated|
      *
      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     P debugNoClose    B                   export
     D debugNoClose    PI

     c                   eval      wkDebugNoClose = *ON

     c                   return
     P                 E


      /define ERRNO_LOAD_PROCEDURE
      /copy QRPGLESRC,ERRNO_H
